from pymxbi.detector import FusionContinuousDetectorModel
from PySide6.QtWidgets import QComboBox, QLabel, QLineEdit

from .....utils.serial import get_all_ports, get_baudrates
from ..device_card import DeviceCard


class FusionDetectorCard(DeviceCard[FusionContinuousDetectorModel]):
    def __init__(self):
        super().__init__()

        self.set_title("Fusion Detector")

        label_pin = QLabel("Beam Break Pin")
        self._line_pin = QLineEdit()
        self._line_pin.setPlaceholderText("Pin Number")
        self.layout_config.addRow(label_pin, self._line_pin)

        label_port = QLabel("Port")
        self._combo_port = QComboBox()
        self._combo_port.addItems(get_all_ports())
        self.layout_config.addRow(label_port, self._combo_port)

        label_baudrate = QLabel("Baudrate")
        self._combo_baudrate = QComboBox()
        for text, value in get_baudrates():
            self._combo_baudrate.addItem(text, value)
        self.layout_config.addRow(label_baudrate, self._combo_baudrate)

        label_poll_interval = QLabel("Poll Interval")
        self._line_poll_interval = QLineEdit()
        self._line_poll_interval.setPlaceholderText("Poll Interval (s)")
        self.layout_config.addRow(label_poll_interval, self._line_poll_interval)

        label_rfid_timeout = QLabel("RFID Timeout")
        self._line_rfid_timeout = QLineEdit()
        self._line_rfid_timeout.setPlaceholderText("RFID Timeout (s)")
        self.layout_config.addRow(label_rfid_timeout, self._line_rfid_timeout)

    def load_config(self, model: FusionContinuousDetectorModel) -> None:
        self.checkbox_enabled.setChecked(model.enabled)
        self.line_device_id.setText(str(model.id))
        self._line_pin.setText(str(model.pin))
        self._combo_port.setCurrentText(model.port)
        self._combo_baudrate.setCurrentText(str(model.baudrate))
        self._line_poll_interval.setText(str(model.poll_interval))
        self._line_rfid_timeout.setText(str(model.rfid_timeout))

    @property
    def result(self) -> FusionContinuousDetectorModel:
        return FusionContinuousDetectorModel(
            enabled=self.checkbox_enabled.isChecked(),
            id=int(self.line_device_id.text()),
            pin=int(self._line_pin.text()),
            port=self._combo_port.currentText(),
            baudrate=int(self._combo_baudrate.currentData()),
            poll_interval=float(self._line_poll_interval.text()),
            rfid_timeout=float(self._line_rfid_timeout.text()),
        )



from __future__ import annotations

from dataclasses import dataclass
from time import time
from tkinter import CENTER, Canvas, Event
from typing import TYPE_CHECKING

import numpy as np
from mxbi.tasks.cross_modal.config import CrossModalConfig
from mxbi.utils.logger import logger
from mxbi.utils.tkinter.components.canvas_with_border import CanvasWithInnerBorder
from mxbi.utils.tkinter.components.showdata_widget import ShowDataWidget
from numpy.typing import NDArray
from PIL import ImageTk

if TYPE_CHECKING:
    from mxbi.models.animal import AnimalState
    from mxbi.models.session import ScreenConfig, SessionState
    from mxbi.tasks.cross_modal.trial_schema import Trial
    from mxbi.theater import Theater
    from PIL.Image import Image as PILImage


@dataclass
class CrossModalResult:
    chosen_side: str | None
    timeout: bool
    feedback: bool
    cancelled: bool
    trial_start_time: float | None
    choice_time: float | None
    choice_x: int | None
    choice_y: int | None


class CrossModalScene:
    def __init__(
        self,
        theater: "Theater",
        session_state: "SessionState",
        animal_state: "AnimalState",
        screen: "ScreenConfig",
        trial: "Trial",
        cross_modal_config: CrossModalConfig,
        left_image: "PILImage",
        right_image: "PILImage",
        audio_stimulus: "NDArray[np.int16]",
    ) -> None:
        self._theater = theater
        self._session_state = session_state
        self._animal_state = animal_state
        self._screen = screen
        self._trial = trial
        self._cross_modal_config = cross_modal_config

        self._left_image_pil = left_image
        self._right_image_pil = right_image
        self._audio_stimulus = audio_stimulus

        self._background: CanvasWithInnerBorder | None = None
        self._left_canvas: Canvas | None = None
        self._right_canvas: Canvas | None = None
        self._left_image: ImageTk.PhotoImage | None = None
        self._right_image: ImageTk.PhotoImage | None = None
        self._show_data_widget: ShowDataWidget | None = None

        self._chosen_side: str | None = None
        self._timeout = False
        self._feedback = False
        self._cancelled = False

        self._trial_start_time: float | None = None
        self._choice_time: float | None = None
        self._choice_x: int | None = None
        self._choice_y: int | None = None

    def start(self) -> CrossModalResult:
        self._create_view()
        self._bind_events()
        self._play_audio()
        self._theater.mainloop()
        return CrossModalResult(
            chosen_side=self._chosen_side,
            timeout=self._timeout,
            feedback=self._feedback,
            cancelled=self._cancelled,
            trial_start_time=self._trial_start_time,
            choice_time=self._choice_time,
            choice_x=self._choice_x,
            choice_y=self._choice_y,
        )

    def cancel(self) -> None:
        if self._cancelled:
            return
        self._cancelled = True
        self._feedback = False
        self._timeout = False
        self._stop_and_close()

    def _create_view(self) -> None:
        self._background = CanvasWithInnerBorder(
            master=self._theater.root,
            bg="black",
            width=self._screen.width,
            height=self._screen.height,
            border_width=40,
        )
        self._background.place(relx=0.5, rely=0.5, anchor="center")
        self._background.focus_set()

        self._show_data_widget = ShowDataWidget(self._background)
        self._show_data_widget.place(relx=0, rely=1, anchor="sw")
        self._show_data_widget.show_data(
            {
                "name": self._animal_state.name,
                "id": self._animal_state.trial_id,
                "level_id": self._animal_state.current_level_trial_id,
                "level": self._animal_state.level,
                "rewards": 0,
                "correct": self._animal_state.correct_trial,
                "incorrect": 0,
                "timeout": 0,
            }
        )

        self._background.create_text(
            self._screen.width / 2,
            self._screen.height / 2,
            text="+",
            fill="white",
            font=("Helvetica", 40),
        )

        fixation_ms = self._cross_modal_config.timing.fixation_ms
        self._background.after(fixation_ms, self._show_images)

    def _show_images(self) -> None:
        if self._background is None:
            return

        self._background.delete("all")

        img_size = int(self._left_image_pil.size[0])

        self._left_canvas = Canvas(
            self._background,
            width=img_size,
            height=img_size,
            bg="grey",
            highlightthickness=0,
        )
        self._left_canvas.place(relx=0.25, rely=0.5, anchor=CENTER)

        self._right_canvas = Canvas(
            self._background,
            width=img_size,
            height=img_size,
            bg="grey",
            highlightthickness=0,
        )
        self._right_canvas.place(relx=0.75, rely=0.5, anchor=CENTER)

        if self._left_canvas is not None:
            self._left_canvas.bind(
                "<ButtonPress-1>", lambda e: self._on_choice("left", e)
            )
        if self._right_canvas is not None:
            self._right_canvas.bind(
                "<ButtonPress-1>", lambda e: self._on_choice("right", e)
            )

        self._left_image = ImageTk.PhotoImage(self._left_image_pil)
        self._left_canvas.create_image(
            img_size // 2, img_size // 2, image=self._left_image
        )

        self._right_image = ImageTk.PhotoImage(self._right_image_pil)
        self._right_canvas.create_image(
            img_size // 2, img_size // 2, image=self._right_image
        )

        self._background.create_text(
            self._screen.width * 0.25,
            self._screen.height * 0.8,
            text=f"Left: {self._trial.left_image_identity_id}",
            fill="white",
        )
        self._background.create_text(
            self._screen.width * 0.75,
            self._screen.height * 0.8,
            text=f"Right: {self._trial.right_image_identity_id}",
            fill="white",
        )

        self._trial_start_time = time()
        timeout_ms = self._cross_modal_config.timing.trial_timeout_ms
        self._background.after(timeout_ms, self._on_timeout)

    def _bind_events(self) -> None:
        if self._background is None:
            return
        self._background.bind("<r>", lambda e: self._give_manual_reward())
        self._background.bind("<s>", lambda e: self._theater.caputre(self._background))

    def _play_audio(self) -> None:
        try:
            self._theater.acontroller.set_master_volume(
                self._cross_modal_config.audio.master_volume
            )
            self._theater.acontroller.set_digital_volume(
                self._cross_modal_config.audio.digital_volume
            )
        except Exception:
            logger.exception("Failed to set cross-modal audio volume")

        self._theater.aplayer.play_stimulus(self._audio_stimulus)

    def _on_choice(self, side: str, event: Event) -> None:
        if self._cancelled or self._chosen_side is not None:
            return
        self._chosen_side = side
        self._choice_time = time()
        self._choice_x = event.x_root
        self._choice_y = event.y_root
        self._timeout = False
        self._feedback = side == self._trial.correct_side
        self._stop_and_close()

    def _on_timeout(self) -> None:
        if self._cancelled or self._chosen_side is not None:
            return
        self._chosen_side = None
        self._timeout = True
        self._feedback = False
        self._stop_and_close()

    def _stop_and_close(self) -> None:
        self._theater.aplayer.stop()
        if self._background is not None:
            try:
                self._background.destroy()
            except Exception:
                pass
            self._background = None
        try:
            self._theater.root.quit()
        except Exception:
            pass

    def _give_manual_reward(self) -> None:
        self._theater.reward.give_reward(500)



from typing import Callable


class EventBus:
    def __init__(self) -> None:
        self._events_dict: dict[str, list[Callable]] = {}

    def subscribe(self, event: str, handler: Callable) -> None:
        if event in self._events_dict:
            return

        self._events_dict[event] = [handler]

    def unsubscribe(self, event: str, handler: Callable) -> None:
        if event not in self._events_dict:
            return

        self._events_dict[event].remove(handler)

    def publish(self, event: str) -> None:
        if event not in self._events_dict:
            return

        for handler in self._events_dict[event]:
            handler()

    def clear(self) -> None:
        self._events_dict.clear()


event_bus = EventBus()



from tkinter import Canvas

from mxbi.utils.tkinter.create_circle import create_circle


class Starter(Canvas):
    def __init__(self, master, size) -> None:
        super().__init__(
            master, width=size, height=size, bg="blue", highlightthickness=0
        )

        create_circle(
            size * 0.5,
            size * 0.5,
            (size - 1) / 2,
            self,
            "white",
        )


if __name__ == "__main__":
    from tkinter import Tk

    root = Tk()
    root.geometry("300x300")
    Starter(root, 300).pack()
    root.mainloop()



from mxbi.utils.logger import logger
from pymotego.send_email import EmailAttachment, EmailClient


def send_email(subject: str, body: str, attachments: list[EmailAttachment]) -> None:
    with EmailClient() as client:
        response = client.send(
            subject=subject,
            html_body=body,
            attachments=attachments,
        )
    logger.info(f"Status: {response.status_code}")
    logger.info(f"Body: {response.text}")



from pathlib import Path

from mxbi.config import Configure
from mxbi.models.animal import ScheduleCondition
from mxbi.tasks.GNGSiD.models import LevelID, MonkeyName
from pydantic import BaseModel, ConfigDict, RootModel

CONFIG_PATH = Path(__file__).parent / "config.json"


class SizeReductionStageParameters(BaseModel):
    model_config = ConfigDict(frozen=True)

    # audio stimulation config
    stimulus_duration: int
    stimulus_freq: int
    stimulus_freq_duration: int
    stimulus_interval: int

    # trial lifecycle
    time_out: int
    inter_trial_interval: int

    # reward config
    reward_duration: int


class SizeReductionStageLeveledParameters(BaseModel):
    model_config = ConfigDict(frozen=True)

    level: int
    stimulation_size: int
    reward_delay: int


class SizeReductionStageConfig(BaseModel):
    model_config = ConfigDict(frozen=True)

    condition: ScheduleCondition

    params: SizeReductionStageParameters
    levels_table: dict[LevelID, SizeReductionStageLeveledParameters]


class SizeReductionStageConfigs(RootModel):
    model_config = ConfigDict(frozen=True)

    root: dict[MonkeyName, SizeReductionStageConfig]


def load_config() -> SizeReductionStageConfigs:
    configs = Configure(CONFIG_PATH, SizeReductionStageConfigs).value
    for config in configs.root.values():
        config.condition.level_count = len(config.levels_table)
    return configs


config = load_config()



from pathlib import Path

from pymxbi import MXBI

from ..infra.timer import FrameTimer
from ..models.session import Session
from .path import get_data_dir_path


class MXBIFlow:
    def __init__(self, session: Session, mxbi: MXBI) -> None:
        self._session = session
        self._timer = FrameTimer()
        self._mxbi = mxbi

    @property
    def session(self) -> Session:
        return self._session

    @property
    def timer(self) -> FrameTimer:
        return self._timer

    @property
    def mxbi(self) -> MXBI:
        return self._mxbi

    @property
    def data_dir(self) -> Path:
        return get_data_dir_path()

    def update(self) -> None:
        self._timer.update()


_current_mxbiflow: MXBIFlow | None = None


def get_mxbiflow() -> MXBIFlow:
    global _current_mxbiflow
    if _current_mxbiflow is None:
        raise RuntimeError("MXBIFlow not initialized")
    return _current_mxbiflow


def set_mxbiflow(mxbiflow: MXBIFlow) -> None:
    global _current_mxbiflow
    if _current_mxbiflow is not None:
        raise RuntimeError("MXBIFlow already initialized")
    _current_mxbiflow = mxbiflow



from pathlib import Path

from mxbi.config import Configure
from mxbi.models.animal import ScheduleCondition
from mxbi.tasks.GNGSiD.models import MonkeyName
from pydantic import BaseModel, ConfigDict, Field, RootModel

CONFIG_PATH = Path(__file__).parent / "config.json"


class InitialHabituationTraingStageLeveledParams(BaseModel):
    model_config = ConfigDict(frozen=True)

    level: int
    entry_reward: float

    evaluation_interval: int

    min_stimulus_interval: float
    max_stimulus_interval: float

    target: float

    stimulus_density: int


class InitialHabituationTrainingStageConfig(BaseModel):
    model_config = ConfigDict(frozen=True)

    condition: ScheduleCondition
    levels_table: dict[int, InitialHabituationTraingStageLeveledParams]


class InitialHabituationTrainingStageConfigs(RootModel):
    model_config = ConfigDict(frozen=True)

    root: dict[MonkeyName, InitialHabituationTrainingStageConfig]


class StageContext(BaseModel):
    duration: int = 0
    rewards: int = 0


class StageContexts(RootModel):
    root: dict[MonkeyName, StageContext] = Field(default_factory=dict)


def load_config() -> InitialHabituationTrainingStageConfigs:
    configs = Configure(CONFIG_PATH, InitialHabituationTrainingStageConfigs).value
    for config in configs.root.values():
        config.condition.level_count = len(config.levels_table)
    return configs


config = load_config()



from .detector.beambreak_detector_card import BeambreakDetectorCard
from .detector.fusion_detector import FusionDetectorCard
from .detector.mock_detector_card import MockDetectorCard
from .detector.rfid_detector import RFIDDetectorCard
from .device_card import DeviceCard
from .rewarder.mock_rewarder_card import MockRewarderCard
from .rewarder.rpi_gpio_rewarder import RPIGpioPumpCard

__all__ = [
    "DeviceCard",
    "MockRewarderCard",
    "RPIGpioPumpCard",
    "MockDetectorCard",
    "RFIDDetectorCard",
    "BeambreakDetectorCard",
    "FusionDetectorCard",
]



from __future__ import annotations

from typing import Generic, TypeVar

from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QIntValidator
from PySide6.QtWidgets import (
    QCheckBox,
    QFormLayout,
    QLabel,
    QLineEdit,
    QMenu,
    QVBoxLayout,
)

from ..card import CardFrame

TModel = TypeVar("TModel")


class DeviceCard(CardFrame, Generic[TModel]):
    remove_requested = Signal()

    def __init__(self, parent=None):
        super().__init__(parent=parent, object_name="card")

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._open_menu)

        self.setMinimumWidth(260)

        self.layout_main = QVBoxLayout()
        self.setLayout(self.layout_main)
        self.layout_main.setContentsMargins(8, 8, 8, 8)

        self.label_title = QLabel("Device Configuration")
        self.layout_main.addWidget(self.label_title)

        self.layout_config = QFormLayout()
        self.layout_main.addLayout(self.layout_config)

        self.lable_enabled = QLabel("Enabled:")
        self.checkbox_enabled = QCheckBox()
        self.checkbox_enabled.setChecked(False)
        self.layout_config.addRow(self.lable_enabled, self.checkbox_enabled)

        self.label_id = QLabel("Device ID:")
        self.line_device_id = QLineEdit("0")
        self.int_validator = QIntValidator(0, 1000, self)
        self.line_device_id.setValidator(self.int_validator)
        self.layout_config.addRow(self.label_id, self.line_device_id)

    def _open_menu(self, position) -> None:
        menu = QMenu(self)
        action_remove = menu.addAction("Remove")
        action_remove.triggered.connect(
            lambda _checked=False: self.remove_requested.emit()
        )
        menu.exec(self.mapToGlobal(position))

    def set_title(self, title: str) -> None:
        self.label_title.setText(title)

    def load_config(self, model: TModel) -> None:
        raise NotImplementedError

    @property
    def result(self) -> TModel:
        raise NotImplementedError



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

_TRIAL_INDEX_CACHE: dict[tuple[Path, str], int] = {}


@dataclass(frozen=True)
class TrialCursor:
    bundle_root: Path
    subject_id: str

    def next_index(self, trial_count: int) -> int:
        if trial_count <= 0:
            raise ValueError("trial_count must be > 0")
        key = (self.bundle_root.resolve(), self.subject_id)
        idx = _TRIAL_INDEX_CACHE.get(key, 0)
        return idx % trial_count

    def advance(self, last_index: int) -> None:
        key = (self.bundle_root.resolve(), self.subject_id)
        _TRIAL_INDEX_CACHE[key] = last_index + 1



from enum import StrEnum, auto

from pydantic import BaseModel


class CrossModalOutcome(StrEnum):
    CORRECT = auto()
    INCORRECT = auto()
    TIMEOUT = auto()
    ABORTED = auto()


class CrossModalResultRecord(BaseModel):
    timestamp: float
    session_id: int | None
    subject_id: str
    partner_id: str

    trial_id: str
    trial_number: int

    call_identity_id: str
    call_category: str
    is_partner_call: bool

    other_identity_id: str
    other_category: str

    partner_side: str
    correct_side: str

    audio_identity_id: str
    audio_index: int
    audio_path: str

    left_image_identity_id: str
    left_image_index: int
    left_image_path: str

    right_image_identity_id: str
    right_image_index: int
    right_image_path: str

    chosen_side: str | None
    chosen_identity_id: str | None

    outcome: CrossModalOutcome

    trial_start_time: float | None
    choice_time: float | None
    latency_sec: float | None

    choice_x: int | None
    choice_y: int | None

    aborted: bool
    timeout: bool



from __future__ import annotations

from collections.abc import Mapping, Sequence
from typing import Generic, TypeVar

from pymxbi.detector import DetectorModel
from pymxbi.rewarder import RewarderModel
from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QDialog,
    QGridLayout,
    QGroupBox,
    QMenu,
    QMessageBox,
    QWidget,
)

from .device_card.device_card import DeviceCard
from .dialog.add_devices_dialog import AddDeviceDialog

T = TypeVar("T", RewarderModel, DetectorModel)


class Devices(QGroupBox, Generic[T]):
    def __init__(
        self,
        parent: QWidget | None,
        title: str,
        *,
        action_label: str,
        device_types: Sequence[str],
        dialog_title: str,
        label: str,
        card_factories: Mapping[str, type[QWidget]],
        columns: int = 2,
    ):
        super().__init__(title, parent)

        self._action_label = action_label
        self._device_types = list(device_types)
        self._dialog_title = dialog_title
        self._label = label
        self._card_factories = dict(card_factories)
        self._columns = max(1, int(columns))

        self._cards: list[DeviceCard[T]] = []

        self._build_ui()
        self._bind_events()

    @property
    def cards(self) -> list[DeviceCard[T]]:
        return list(self._cards)

    def results(self) -> list[T]:
        return [card.result for card in self._cards]

    def load_models(self, models: Sequence[T]) -> None:
        for model in models:
            self.add_model(model)

    def add_model(self, model: T) -> None:
        try:
            device_type = model.device_type
        except Exception as exc:
            QMessageBox.warning(
                self,
                "Unsupported device",
                f"Failed to determine device type ({exc!s})",
            )
            return
        if device_type is None:
            QMessageBox.warning(
                self,
                "Unsupported device",
                "Missing device type",
            )
            return

        self._add_device_card(str(device_type), model=model)

    # -----------------------------
    # UI / Events
    # -----------------------------

    def _build_ui(self) -> None:
        layout = QGridLayout()
        layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.setLayout(layout)

    def _bind_events(self) -> None:
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._open_menu)

    def _open_menu(self, position) -> None:
        menu = QMenu(self)
        action = menu.addAction(self._action_label)
        action.triggered.connect(lambda _checked=False: self._on_add_clicked())
        menu.exec(self.mapToGlobal(position))

    def _on_add_clicked(self) -> None:
        dialog = AddDeviceDialog(
            self,
            device_types=self._device_types,
            title=self._dialog_title,
            label=self._label,
        )
        if dialog.exec() != QDialog.DialogCode.Accepted:
            return

        self._add_device_card(dialog.device_type, model=None)

    # -----------------------------
    # Cards / Layout
    # -----------------------------

    def _add_device_card(self, device_type: str, *, model: T | None) -> None:
        card_factory = self._card_factories.get(device_type)
        if card_factory is None:
            QMessageBox.warning(
                self,
                "Unsupported device",
                f"Unsupported device type: {device_type}",
            )
            return

        card = card_factory()
        if isinstance(card, DeviceCard):
            if model is not None:
                card.load_config(model)
            card.remove_requested.connect(lambda c=card: self._on_remove_requested(c))
            self._cards.append(card)

        self._mount(card)

    def _on_remove_requested(self, card: DeviceCard[T]) -> None:
        self._remove_card(card)

    def _remove_card(self, card: DeviceCard[T]) -> None:
        layout = self.layout()
        if isinstance(layout, QGridLayout):
            layout.removeWidget(card)
        card.setParent(None)
        card.deleteLater()

        self._reflow_cards()

    def _reflow_cards(self) -> None:
        layout = self.layout()
        if not isinstance(layout, QGridLayout):
            return

        cards: list[DeviceCard[T]] = []
        for i in range(layout.count()):
            item = layout.itemAt(i)
            if item is None:
                continue
            widget = item.widget()
            if isinstance(widget, DeviceCard):
                cards.append(widget)

        while layout.count() > 0:
            layout.takeAt(0)

        self._cards = cards
        for index, widget in enumerate(cards):
            row, col = divmod(index, self._columns)
            layout.addWidget(widget, row, col)

        self._update_column_stretch(layout)

    def _mount(self, widget: QWidget) -> None:
        layout = self.layout()
        if layout is None:
            return

        if isinstance(layout, QGridLayout):
            index = layout.count()
            row, col = divmod(index, self._columns)
            layout.addWidget(widget, row, col)
            self._update_column_stretch(layout)
            return

        layout.addWidget(widget)

    def _update_column_stretch(self, layout: QGridLayout) -> None:
        used_cols = min(len(self._cards), self._columns)
        for col in range(self._columns):
            layout.setColumnStretch(col, 1 if col < used_cols else 0)



from PySide6.QtCore import Signal
from PySide6.QtWidgets import (
    QHBoxLayout,
    QMainWindow,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from ..core.config_store import ConfigStore
from ..core.path import get_config_session_path, get_options_session_path
from ..models.session import Options, SessionConfig
from ..scene import SceneManager
from .components.experiment_groups import (
    ExperimentAnimalsGroup,
    ExperimentConfigGroup,
    ExperimentSceneGroup,
)


class ExperimentPanel(QMainWindow):
    accepted = Signal()
    rejected = Signal()

    def __init__(self, scene_manager: SceneManager):
        super().__init__()
        self._accepted = False
        self._config = ConfigStore(get_config_session_path(), SessionConfig)
        self._options = ConfigStore(get_options_session_path(), Options)
        self._stage_level_tables = scene_manager.stage_level_tables

        self.setWindowTitle("Experiment Panel")

        central_widget = QWidget(self)
        self.setCentralWidget(central_widget)

        layout_main = QVBoxLayout(self)
        central_widget.setLayout(layout_main)

        self.group_config = ExperimentConfigGroup(
            self, self._options.value.experimenter
        )
        layout_main.addWidget(self.group_config)

        self.group_scene = ExperimentSceneGroup(
            self, stages=list(self._stage_level_tables.keys())
        )
        layout_main.addWidget(self.group_scene)

        self.group_animals = ExperimentAnimalsGroup(
            self,
            animals=self._options.value.animals,
            stage_level_tables=self._stage_level_tables,
        )
        layout_main.addWidget(self.group_animals)

        self.combo_experimenter = self.group_config.combo_experimenter
        self.combo_reward_type = self.group_config.combo_reward_type
        self.line_notes = self.group_config.line_notes

        layout_buttons = QHBoxLayout(self)
        layout_main.addLayout(layout_buttons)
        self._button_cancle = QPushButton("Cancel", self)
        self._button_save = QPushButton("Save", self)
        self._button_continue = QPushButton("Continue", self)
        layout_buttons.addWidget(self._button_cancle)
        layout_buttons.addWidget(self._button_save)
        layout_buttons.addWidget(self._button_continue)

        self._bind_signals()
        self.load_config()

    def _bind_signals(self):
        self._button_cancle.clicked.connect(self._on_cancle)
        self._button_save.clicked.connect(self._on_save)
        self._button_continue.clicked.connect(self._on_continue)

    def load_config(self):
        self.group_config.load_config(self._config.value)
        self.group_scene.load_config(self._config.value)
        self.group_animals.load_config(self._config.value)

    def result(self) -> SessionConfig:
        session_config = self.group_config.result()
        scene_result = self.group_scene.result()
        session_config.default_scene = scene_result.default_scene
        session_config.unknown_animal_fallback = scene_result.unknown_animal_fallback
        session_config.fault_fallback = scene_result.fault_fallback
        session_config.animals = self.group_animals.result()
        return session_config

    def _on_save(self):
        self._config.save(self.result())

    def _on_cancle(self):
        self.close()

    def _on_continue(self):
        self._on_save()
        self._accepted = True
        self.close()
        self.accepted.emit()

    def closeEvent(self, event) -> None:
        super().closeEvent(event)
        if event.isAccepted() and not self._accepted:
            self.rejected.emit()



import pymxbi
from loguru import logger
from pymxbi import MXBIModel

from .core.config_store import ConfigStore
from .core.path import (
    get_config_session_path,
    get_mxbi_config_path,
    get_session_counter_path,
)
from .gameloop.detector_bridge import DetectorBridge
from .gameloop.game import Game
from .models.animal import Animal, StageState
from .models.session import DailySessionIdStore, Session, SessionConfig
from .scene import SceneManager


def init_gameloop(scene_manager: SceneManager) -> Game:
    """
    Initialize the game loop with MXBI, session, and detector bridge.

    Parameters
    ----------
    scene_manager : SceneManager
        The scene manager instance to be used by the game.

    Returns
    -------
    Game
        The initialized game instance ready to run.

    Notes
    -----
    This function orchestrates the initialization of all core components
    required for the MXBI game loop to function.
    """
    mxbi = build_mxbi()
    session_config = ConfigStore(get_config_session_path(), SessionConfig).value
    session = init_session(session_config)

    detector_bridge = DetectorBridge(
        mxbi.detector, {i.rfid_id: i.name for i in session.animals.values()}
    )

    return Game(session, scene_manager, detector_bridge, mxbi)


def build_mxbi() -> pymxbi.MXBI:
    mxbi_config = ConfigStore(get_mxbi_config_path(), MXBIModel).value
    return pymxbi.build_mxbi(mxbi_config, logger)


def init_session(session_config: SessionConfig) -> Session:
    store = DailySessionIdStore(get_session_counter_path())

    animal_dict: dict[str, Animal] = {}
    for animal_config in session_config.animals:
        train_state = StageState(
            stage_name=animal_config.stage, level=animal_config.level
        )
        animal_state = Animal(
            rfid_id=animal_config.rfid_id,
            name=animal_config.name,
        )
        animal_state.set_current_stage(train_state)
        animal_dict[animal_config.name] = animal_state

    session = Session(
        session_id=store.session_id,
        experimenter=session_config.experimenter,
        reward_type=session_config.reward_type,
        send_email=False,
        sync_data=False,
        note=session_config.note,
        default_scene=session_config.default_scene,
        unknown_animal_fallback=session_config.unknown_animal_fallback,
        fault_fallback=session_config.fault_fallback,
        hide_cursor=session_config.hide_cursor,
        fullscreen=session_config.fullscreen,
        animals=animal_dict,
    )
    session.start()

    return session



import os
import tempfile
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path

from pydantic import BaseModel, Field, PrivateAttr, ValidationError

from .animal import Animal, AnimalConfig
from .reward import RewardEnum


class SessionConfig(BaseModel):
    experimenter: str = Field(default="auto", frozen=True)
    reward_type: RewardEnum = Field(default=RewardEnum.AGUM_ONE_FIFTH, frozen=True)
    send_email: bool = Field(default=False, frozen=True)
    sync_data: bool = Field(default=False, frozen=True)
    note: str = Field(default="", max_length=1000)

    default_scene: str = Field(default="")
    unknown_animal_fallback: str = Field(default="")
    fault_fallback: str = Field(default="")

    hide_cursor: bool = Field(default=False)
    fullscreen: bool = Field(default=False)

    animals: list[AnimalConfig] = Field(default_factory=list)


class DailySessionCounter(BaseModel):
    day: str = Field(
        default_factory=lambda: datetime.now(timezone.utc).date().isoformat()
    )
    last_session_id: int = Field(default=0, ge=0)


@dataclass
class DailySessionIdStore:
    path: Path

    def _today_utc(self):
        return datetime.now(timezone.utc).date().isoformat()

    def _load(self) -> DailySessionCounter:
        if not self.path.exists():
            return DailySessionCounter()

        text = self.path.read_text()

        try:
            return DailySessionCounter.model_validate_json(text)
        except ValueError, ValidationError:
            return DailySessionCounter()

    def _atomic_write(self, data: DailySessionCounter) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        fd, tmp = tempfile.mkstemp(
            dir=str(self.path.parent),
            prefix=self.path.name,
            suffix=".tmp",
        )

        try:
            with os.fdopen(fd, "w", encoding="utf-8") as f:
                f.write(data.model_dump_json())
                f.flush()
                os.fsync(f.fileno())

            os.replace(tmp, self.path)
        finally:
            try:
                os.remove(tmp)
            except FileNotFoundError:
                pass

    @property
    def session_id(self) -> int:
        today = self._today_utc()
        data = self._load()

        if data.day != today:
            data.day = today
            data.last_session_id = 0

        data.last_session_id += 1
        self._atomic_write(data)

        return data.last_session_id


class Session(BaseModel):
    experimenter: str = Field(default="auto", frozen=True)
    reward_type: RewardEnum = Field(default=RewardEnum.AGUM_ONE_FIFTH, frozen=True)
    send_email: bool = Field(default=False, frozen=True)
    sync_data: bool = Field(default=False, frozen=True)

    default_scene: str = Field(default="", frozen=True)
    unknown_animal_fallback: str = Field(default="", frozen=True)
    fault_fallback: str = Field(default="", frozen=True)

    hide_cursor: bool = Field(default=False, frozen=True)
    fullscreen: bool = Field(default=False, frozen=True)

    session_id: int = Field(default=0, ge=0)
    start_at: float = Field(default=0, ge=0)
    end_at: float = Field(default=0, ge=0)
    note: str = Field(frozen=True)

    _current_animal: str | None = PrivateAttr(default=None)
    animals: dict[str, Animal] = Field(default_factory=dict, frozen=True)

    def start(self):
        self.start_at = datetime.now(timezone.utc).timestamp()

    def end(self):
        self.end_at = datetime.now(timezone.utc).timestamp()

    @property
    def current_animal(self) -> Animal | None:
        key = self._current_animal
        if key is None:
            return None

        return self.animals[key]

    @property
    def require_current_animal(self) -> Animal:
        key = self._current_animal
        if key is None:
            raise RuntimeError(
                f"Animal: {key} is not set. Please call set_current_animal() first"
            )

        return self.animals[key]

    def clear_current_animal(self):
        a = self.current_animal
        if a is None:
            return

        a.end_animal_session()
        self._current_animal = None

    def set_current_animal(self, animal: str):
        self.clear_current_animal()

        try:
            a = self.animals[animal]
        except KeyError:
            raise ValueError(f"animal {animal} not found")

        self._current_animal = animal
        a.start_animal_session()


class Options(BaseModel):
    mxbis: list[str] = Field(default_factory=list, frozen=True)
    experimenter: list[str] = Field(default_factory=list, frozen=True)
    animals: dict[str, str] = Field(default_factory=dict, frozen=True)



from tkinter import Canvas


class BaseViews:
    def _create_background(self, master, width, height) -> None:
        self._backgroud = Canvas(
            master,
            bg="black",
            width=width,
            height=height,
            highlightthickness=0,
        )
        self._backgroud.place(relx=0.5, rely=0.5, anchor="center")



from tkinter import Canvas


def create_circle(
    x_coord: float, y_coord: float, r: float, canvas: Canvas, color: str
) -> int:
    x0 = x_coord - r
    y0 = y_coord - r
    x1 = x_coord + r
    y1 = y_coord + r
    return canvas.create_oval(x0, y0, x1, y1, outline="", fill=color)


class DetectTarget(Canvas):
    def __init__(self, master, size) -> None:
        super().__init__(
            master, width=size, height=size, bg="lightgray", highlightthickness=0
        )
        # TODO: figure out magic number
        circle_config = [
            (0.5, 2.1, "#616161"),
            (0.5, 3.1, "#bababa"),
            (0.5, 6.3, "white"),
        ]
        for cx, divisor, color in circle_config:
            create_circle(
                size * cx,
                size * cx,
                size / divisor,
                self,
                color,
            )


class DiscriminateTarget(Canvas):
    def __init__(self, master, size) -> None:
        super().__init__(
            master, width=size, height=size, bg="lightgray", highlightthickness=0
        )

        circle_config = [
            (0.5, 0.5, 3.1, "#616161"),
            (0.5, 0.5, 6.3, "white"),
            (0.25, 0.25, 6.3, "#616161"),
            (0.75, 0.25, 6.3, "#616161"),
            (0.25, 0.75, 6.3, "#616161"),
            (0.75, 0.75, 6.3, "#616161"),
        ]

        for cx, cy, divisor, color in circle_config:
            create_circle(
                size * cx,
                size * cy,
                size / divisor,
                self,
                color,
            )



from pymxbi.rewarder import GPIORewarderModel
from PySide6.QtGui import QIntValidator
from PySide6.QtWidgets import (
    QLabel,
    QLineEdit,
)

from ..device_card import DeviceCard


class RPIGpioPumpCard(DeviceCard[GPIORewarderModel]):
    def __init__(self):
        super().__init__()
        self.set_title("GPIO Pump")

        lable_gpio_pin = QLabel("GPIO Pin:")
        self.line_gpio_pin = QLineEdit("0")
        int_validator = QIntValidator(0, 40, self)
        self.line_gpio_pin.setValidator(int_validator)
        self.layout_config.addRow(lable_gpio_pin, self.line_gpio_pin)

    def load_config(self, model: GPIORewarderModel) -> None:
        self.checkbox_enabled.setChecked(model.enabled)
        self.line_device_id.setText(str(model.id))
        self.line_gpio_pin.setText(str(model.pin))

    @property
    def result(self) -> GPIORewarderModel:
        return GPIORewarderModel(
            enabled=self.checkbox_enabled.isChecked(),
            id=int(self.line_device_id.text()),
            pin=int(self.line_gpio_pin.text()),
        )



from pathlib import Path

from mxbi.config import Configure
from mxbi.models.animal import ScheduleCondition
from mxbi.tasks.two_alternative_choice.models import LevelID, MonkeyName
from mxbi.tasks.two_alternative_choice.tasks.touch.touch_models import TrialConfig
from pydantic import BaseModel, ConfigDict, RootModel

CONFIG_PATH = Path(__file__).parent / "config.json"


class SizeReductionStageLeveledParameters(BaseModel):
    model_config = ConfigDict(frozen=True)

    level: int
    stimulation_size: int


class SizeReductionStageConfig(BaseModel):
    model_config = ConfigDict(frozen=True)

    condition: ScheduleCondition
    trial_config: TrialConfig

    levels_table: dict[LevelID, SizeReductionStageLeveledParameters]


class SizeReductionStageConfigs(RootModel):
    model_config = ConfigDict(frozen=True)

    root: dict[MonkeyName, SizeReductionStageConfig]


def load_config() -> SizeReductionStageConfigs:
    configs = Configure(CONFIG_PATH, SizeReductionStageConfigs).value
    for config in configs.root.values():
        config.condition.level_count = len(config.levels_table)
    return configs


config = load_config()



import json
from pathlib import Path
from typing import Literal

from pydantic import BaseModel, Field

from ...path import get_cross_modal_config_path


class CrossModalVisualConfig(BaseModel):
    image_scale: float = Field(default=0.5, ge=0.1, le=0.9)


class CrossModalAudioConfig(BaseModel):
    master_volume: int = Field(default=70, ge=0, le=100)
    digital_volume: int = Field(default=70, ge=0, le=100)
    gain: float = Field(default=1.0, ge=0.0, le=4.0)
    wav_rate_policy: Literal["resample", "error"] = "resample"


class CrossModalTimingConfig(BaseModel):
    fixation_ms: int = Field(default=300, ge=0, le=10_000)
    trial_timeout_ms: int = Field(default=10_000, ge=1_000, le=120_000)


class CrossModalConfig(BaseModel):
    visual: CrossModalVisualConfig = Field(default_factory=CrossModalVisualConfig)
    audio: CrossModalAudioConfig = Field(default_factory=CrossModalAudioConfig)
    timing: CrossModalTimingConfig = Field(default_factory=CrossModalTimingConfig)


def load_cross_modal_config(path: Path | None = None) -> CrossModalConfig:
    if path is None:
        path = get_cross_modal_config_path()
    if not path.exists():
        raise FileNotFoundError(f"Cross-modal config not found: {path}")
    return CrossModalConfig.model_validate_json(path.read_text(encoding="utf-8"))


def save_cross_modal_config(config: CrossModalConfig, path: Path | None = None) -> None:
    if path is None:
        path = get_cross_modal_config_path()
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(json.dumps(config.model_dump(), indent=4), encoding="utf-8")



from dataclasses import dataclass
from pathlib import Path
from random import choice

from pygame import Event, Rect, Surface, image, transform

from mxbiflow import get_mxbiflow

ASSETS_PATH = Path(__file__).parent / "assets"


@dataclass
class Asset:
    image: Surface
    rect: Rect


class IDLE:
    _running: bool
    level_table: dict[str, list[int]] = {}

    def __init__(self) -> None:
        self._mxbiflow = get_mxbiflow()

        self._screen_size = self._mxbiflow.mxbi.screen_size
        self._pos = ((self._screen_size.width // 4) * 3, self._screen_size.height // 2)
        self._vstimulus_size = self._screen_size.width // 2 * 0.75

        self._assets = [
            Asset(
                asset_image := transform.scale(
                    image.load(path).convert_alpha(),
                    (self._vstimulus_size, self._vstimulus_size),
                ),
                asset_image.get_rect(center=self._pos),
            )
            for path in ASSETS_PATH.glob("*.png")
        ]

        self._asset = choice(self._assets)

    def start(self) -> None:
        self._running = True

    def quit(self) -> None:
        self._running = False

    @property
    def running(self) -> bool:
        return self._running

    def handle_event(self, event: Event) -> None: ...

    def update(self, dt_s: float) -> None: ...

    def draw(self, screen: Surface) -> None:
        screen.blit(self._asset.image, self._asset.rect)



from dataclasses import dataclass
from typing import Callable

from pygame import time


@dataclass
class _Task:
    interval_ms: int
    callback: Callable[[], None]
    repeat: bool
    next_fire_ms: int
    cancelled: bool = False

    def cancel(self) -> None:
        self.cancelled = True


class FrameTimer:
    def __init__(self) -> None:
        self._tasks: list[_Task] = []

    def after(self, delay_ms: int, callback: Callable[[], None]) -> _Task:
        now = time.get_ticks()
        task = _Task(delay_ms, callback, False, now + delay_ms)
        self._tasks.append(task)
        return task

    def every(self, interval_ms: int, callback: Callable[[], None]) -> _Task:
        now = time.get_ticks()
        task = _Task(interval_ms, callback, True, now + interval_ms)
        self._tasks.append(task)
        return task

    def cancel(self, task: _Task) -> None:
        task.cancel()

    def cancel_all(self) -> None:
        for task in self._tasks:
            task.cancel()
        self._tasks.clear()

    def update(self) -> None:
        now = time.get_ticks()
        for task in self._tasks:
            if task.cancelled:
                continue
            if now >= task.next_fire_ms:
                task.callback()
                if task.repeat and not task.cancelled:
                    task.next_fire_ms += task.interval_ms
                else:
                    task.cancel()

        self._tasks = [task for task in self._tasks if not task.cancelled]



from pathlib import Path

_base_path: Path | None = None

CONFIG_SESSION_FILENAME = "session.json"
OPTIONS_SESSION_FILENAME = "options.json"
MXBI_CONFIG_FILENAME = "mxbi.json"
CROSS_MODAL_CONFIG_FILENAME = "config_cross_modal.json"
SAMBA_BACKUP_DIR_NAME = "backup"
SERVICE_DIR_NAME = "services"
MOUNT_SERVICE_NAME = "mount.service"
SYNC_SERVICE_NAME = "sync.service"


def set_base_path(path: Path | str) -> None:
    global _base_path
    _base_path = Path(path)


def get_base_path() -> Path:
    if _base_path is None:
        raise RuntimeError("Base path not set. Call set_base_path() first.")
    return _base_path


def get_config_dir_path() -> Path:
    return get_base_path() / "config"


def get_mxbi_config_path() -> Path:
    return get_config_dir_path() / MXBI_CONFIG_FILENAME


def get_config_session_path() -> Path:
    return get_config_dir_path() / CONFIG_SESSION_FILENAME


def get_options_session_path() -> Path:
    return get_config_dir_path() / OPTIONS_SESSION_FILENAME


def get_cross_modal_config_path() -> Path:
    return get_config_dir_path() / CROSS_MODAL_CONFIG_FILENAME


def get_data_dir_path() -> Path:
    return get_base_path() / "data"


def get_log_path() -> Path:
    return get_base_path() / "log"


def get_samba_mount_path() -> Path:
    return get_base_path() / "samba_mount"


def get_samba_backup_dir_path() -> Path:
    return get_samba_mount_path() / SAMBA_BACKUP_DIR_NAME


def get_service_dir_path() -> Path:
    return get_base_path() / SERVICE_DIR_NAME


def get_mount_service_path() -> Path:
    return get_service_dir_path() / MOUNT_SERVICE_NAME


def get_sync_service_path() -> Path:
    return get_service_dir_path() / SYNC_SERVICE_NAME


def get_internal_state_path() -> Path:
    return get_base_path() / ".mxbiflow" / "state"


def get_session_counter_path() -> Path:
    return get_internal_state_path() / "session_counter.json"



from mxbi.tasks.GNGSiD.models import BaseDataToShow, BaseTrialConfig, BaseTrialData


class TrialConfig(BaseTrialConfig):
    stimulus_freq: int
    stimulus_freq_duration: int
    stimulus_freq_master_amp: int
    stimulus_freq_digital_amp: int

    stimulus_interval: int


class TrialData(BaseTrialData):
    trial_config: TrialConfig


class DataToShow(BaseDataToShow): ...



from .data_logger import DataLogger, DataLoggerType
from .eventbus import EventBus, event_bus
from .flow import Flow
from .timer import FrameTimer

__all__ = [
    "DataLogger",
    "DataLoggerType",
    "EventBus",
    "event_bus",
    "Flow",
    "FrameTimer",
]



import sys

from loguru import logger

from ..core.path import get_log_path

logger.remove()

logger.add(sys.stderr, level="DEBUG")


def init_logger() -> None:
    logger.add(
        f"{get_log_path()}/mxbi.log",
        rotation="10 MB",
        retention="7 days",
        compression="zip",
        encoding="utf-8",
        level="DEBUG",
        serialize=True,
    )



from pathlib import Path
from tkinter.ttk import Label
from typing import TYPE_CHECKING

from mxbi.utils.tkinter.components.canvas_with_border import CanvasWithInnerBorder
from PIL import Image, ImageTk

if TYPE_CHECKING:
    from mxbi.models.animal import AnimalState, ScheduleCondition
    from mxbi.models.session import SessionState
    from mxbi.models.task import Feedback
    from mxbi.theater import Theater

ASSETS_PATH = Path(__file__).parent / "assets"


class IDLEScene:
    def __init__(
        self,
        theater: "Theater",
        session_state: "SessionState",
        animal_state: "AnimalState",
    ) -> None:
        self._theater = theater
        self._session_config = session_state
        self._screen_type = self._session_config.session_config.screen_type
        self._standard_reward_stimulus = self._theater.new_standard_reward_stimulus(
            1000
        )

        self._on_trial_start()

    def start(self) -> "Feedback":
        self._theater.root.mainloop()

        return True

    def _on_trial_start(self) -> None:
        self._create_view()
        self._bind_events()

    def _on_trial_end(self) -> None:
        self._theater.aplayer.stop()
        self._background.destroy()
        self._theater.root.quit()

    def _create_view(self) -> None:
        self._background = CanvasWithInnerBorder(
            master=self._theater.root,
            bg="black",
            width=self._screen_type.width,
            height=self._screen_type.height,
            border_width=40,
        )
        self._background.place(relx=0.5, rely=0.5, anchor="center")

        xshift = 240
        xcenter = self._screen_type.width / 2 + xshift
        ycenter = self._screen_type.height / 2

        self._img = Image.open(ASSETS_PATH / "apple_v1.png")
        self._img = self._img.resize((400, 400)).rotate(-90, expand=True)
        self._img = ImageTk.PhotoImage(self._img)
        self.label_apple = Label(self._background, image=self._img)
        self.label_apple.place(x=xcenter, y=ycenter, anchor="center")

    def _bind_events(self) -> None:
        self._background.focus_set()
        self._background.bind("<r>", lambda e: self._give_stimulus(1000))

    def _give_stimulus(self, duration: int) -> None:
        self._standard_reward_stimulus.play(1000)

    def quit(self) -> None:
        self._on_trial_end()

    def on_idle(self) -> None:
        self._on_trial_end()

    def on_return(self) -> None:
        self._on_trial_end()

    @property
    def condition(self) -> "ScheduleCondition | None":
        return None



from dataclasses import dataclass
from queue import Empty, SimpleQueue

import pygame
from pygame import Event, event
from pymxbi.detector import MockDetector
from pymxbi.detector.detector import DetectionResult, Detector, DetectorEvent

EVT_DETECTOR = pygame.USEREVENT + 1


@dataclass(frozen=True)
class DetectorMsg:
    kind: DetectorEvent
    animal: str | None


class DetectorBridge:
    def __init__(
        self,
        detector: Detector,
        animals_map: dict[str, str],
    ) -> None:
        self._detector = detector
        self._animals_map = animals_map
        self._q: SimpleQueue[DetectorMsg] = SimpleQueue()
        self._started = False

    def _resolve_animal(self, rfid_id: str | None) -> str | None:
        if rfid_id is None:
            return None
        return self._animals_map.get(rfid_id)

    def start(self) -> None:
        if self._started:
            return
        self._started = True

        self._detector.begin()

        # fmt: off
        self._detector.register_event(DetectorEvent.ANIMAL_ENTERED, self._on_animal_entered)
        self._detector.register_event(DetectorEvent.ANIMAL_LEFT, self._on_animal_left)
        self._detector.register_event(DetectorEvent.UNKNOWN_ANIMAL_ENTERED, self._on_unknown_animal_entered)
        self._detector.register_event(DetectorEvent.FAULT_DETECTED, self._on_fault_detected)
        # fmt: on

    def _enqueue(self, kind: DetectorEvent, animal: str | None) -> None:
        self._q.put(DetectorMsg(kind=kind, animal=animal))

    def _on_animal_entered(self, result: DetectionResult) -> None:
        name = self._resolve_animal(result.animal_id)
        if name is None:
            self._enqueue(DetectorEvent.UNKNOWN_ANIMAL_ENTERED, result.animal_id)
            return
        self._enqueue(DetectorEvent.ANIMAL_ENTERED, name)

    def _on_animal_left(self, result: DetectionResult) -> None:
        self._enqueue(DetectorEvent.ANIMAL_LEFT, self._resolve_animal(result.animal_id))

    def _on_unknown_animal_entered(self, result: DetectionResult) -> None:
        self._enqueue(DetectorEvent.UNKNOWN_ANIMAL_ENTERED, result.animal_id)

    def _on_fault_detected(self, result: DetectionResult) -> None:
        self._enqueue(DetectorEvent.FAULT_DETECTED, result.animal_id)

    def emit_pygame_event(self) -> None:
        while True:
            try:
                msg = self._q.get_nowait()
            except Empty:
                break

            event.post(event.Event(EVT_DETECTOR, msg=msg))

    def manual_emit(self, animal_id: str | None = None) -> None:
        if isinstance(self._detector, MockDetector):
            if animal_id is None:
                self._detector.animal_left()
            else:
                self._detector.animal_present(animal_id)

    def _rfid_key_at(self, index: int) -> str | None:
        keys = list(self._animals_map.keys())
        if index < len(keys):
            return keys[index]
        return None

    def handle_event(self, event: Event) -> None:
        if event.type == pygame.KEYDOWN:
            key_map = {
                pygame.K_0: 0,
                pygame.K_1: 1,
                pygame.K_2: 2,
                pygame.K_3: 3,
                pygame.K_4: 4,
                pygame.K_5: 5,
            }
            if event.key in key_map:
                rfid_key = self._rfid_key_at(key_map[event.key])
                if rfid_key is not None:
                    self.manual_emit(rfid_key)
            elif event.key == pygame.K_l:
                self.manual_emit()



from pymxbi.rewarder import MockRewarderModel

from ..device_card import DeviceCard


class MockRewarderCard(DeviceCard[MockRewarderModel]):
    def __init__(self):
        super().__init__()
        self.set_title("Mock Rewarder")

    def load_config(self, model: MockRewarderModel) -> None:
        self.checkbox_enabled.setChecked(model.enabled)
        self.line_device_id.setText(str(model.id))

    @property
    def result(self) -> MockRewarderModel:
        return MockRewarderModel(
            enabled=self.checkbox_enabled.isChecked(),
            id=int(self.line_device_id.text()),
        )



from serial import SerialBase
from serial.tools import list_ports


def get_baudrates():
    return [(str(baudrate), baudrate) for baudrate in SerialBase.BAUDRATES]


def get_all_ports():
    return [port.device for port in list_ports.comports() if port.device is not None]



from mxbi.tasks.two_alternative_choice.models import (
    BaseDataToShow,
    BaseTrialConfig,
    BaseTrialData,
)


class TrialConfig(BaseTrialConfig):
    stimulus_freq: int
    stimulus_freq_duration: int

    stimulus_interval: int


class TrialData(BaseTrialData):
    trial_config: TrialConfig


class DataToShow(BaseDataToShow): ...



import csv
import json
import sys
from datetime import datetime
from enum import StrEnum
from pathlib import Path

from ..utils.logger import logger

now = datetime.now()


class DataLoggerType(StrEnum):
    JSONL = "jsonl"
    JSON = "json"
    CSV = "csv"


class DataLogger:
    def __init__(
        self,
        path: Path,
        session_id: int,
        monkey: str,
        filename: str,
        type: DataLoggerType = DataLoggerType.JSONL,
    ) -> None:
        self._path = path
        self._session_id = session_id
        self._monkey = monkey
        self._filename = filename
        self._type = type

        self._data_dir = self._ensure_data_dir()
        self._data_path = self._get_path(f".{self._type.value}")

    @property
    def path(self) -> Path:
        return self._data_path

    def _ensure_data_dir(self) -> Path:
        date_path = Path(f"{now.year}{now.month:02d}{now.day:02d}")
        session_path = Path(f"{self._session_id}")
        monkey_path = Path(f"{self._monkey}")

        base_dir = self._path / date_path / session_path / monkey_path

        try:
            base_dir.mkdir(parents=True, exist_ok=True)
            return base_dir
        except Exception as e:
            logger.error(f"failed to create {base_dir}: {e}")
            sys.exit(1)

    def _get_path(self, suffix: str) -> Path:
        return self._data_dir / f"{self._filename}{suffix}"

    def save(self, data: dict) -> None:
        match self._type:
            case DataLoggerType.JSONL:
                self._save_jsonl(data)
            case DataLoggerType.JSON:
                self._save_json(data)
            case DataLoggerType.CSV:
                self.save_csv_row(data)

    def _save_jsonl(self, data: dict) -> None:
        try:
            json_line = json.dumps(data, ensure_ascii=False)

            with open(self._data_path, "a", encoding="utf-8") as f:
                f.write(json_line + "\n")

        except TypeError as e:
            logger.error(f"Data is not JSON serializable: {e}")
            raise
        except IOError as e:
            logger.error(f"Failed to write to file {self._data_path}: {e}")
            raise
        except Exception as e:
            logger.error(f"Unexpected error while writing data: {e}")
            raise

    def _save_json(self, data: dict) -> None:
        try:
            self._data_path.parent.mkdir(parents=True, exist_ok=True)
            with open(self._data_path, "w", encoding="utf-8") as f:
                json.dump(data, f, ensure_ascii=False, indent=2)
        except TypeError as e:
            logger.error(f"Data is not JSON serializable: {e}")
            raise
        except IOError as e:
            logger.error(f"Failed to write to file {self._data_path}: {e}")
            raise
        except Exception as e:
            logger.error(f"Unexpected error while writing JSON data: {e}")
            raise

    def save_csv_row(self, data: dict) -> None:
        csv_path = self._get_path(".csv")
        try:
            csv_path.parent.mkdir(parents=True, exist_ok=True)
            file_exists = csv_path.exists() and csv_path.stat().st_size > 0

            fieldnames = sorted(data.keys())

            with csv_path.open("a", newline="", encoding="utf-8") as f:
                writer = csv.DictWriter(f, fieldnames=fieldnames)
                if not file_exists:
                    writer.writeheader()
                writer.writerow({k: data.get(k, "") for k in fieldnames})
        except Exception as e:
            logger.error(f"Failed to write CSV row to {csv_path}: {e}")
            raise



from dataclasses import dataclass, field
from typing import Callable, Generic, List, Protocol, TypeVar

T = TypeVar("T")


class Next(Protocol):
    def __call__(self) -> None: ...


class Step(Protocol):
    def __call__(self, next_t: Next) -> None: ...


@dataclass
class Flow(Generic[T]):
    ctx: T
    _steps: List[Step] = field(default_factory=list)
    _i: int = 0
    _stopped: bool = False

    @classmethod
    def start(cls, ctx: T) -> Flow[T]:
        return cls(ctx=ctx, _steps=[])

    def do(self, fn: Callable[[T], None]) -> Flow[T]:
        def _step(next_t: Next) -> None:
            fn(self.ctx)
            next_t()

        self._steps.append(_step)
        return self

    def then(self, step: Step) -> Flow:
        self._steps.append(step)
        return self

    def run(self) -> None:
        self._next()

    def stop(self) -> None:
        self._stopped = True

    def _next(self) -> None:
        if self._stopped or self._i >= len(self._steps):
            return

        step = self._steps[self._i]
        self._i += 1

        called = False

        def once() -> None:
            nonlocal called
            if called or self._stopped:
                return

            called = True
            self._next()

        step(once)



from tkinter import Canvas
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mxbi.models.animal import AnimalState, ScheduleCondition
    from mxbi.models.session import SessionState
    from mxbi.models.task import Feedback
    from mxbi.theater import Theater


class ErrorScene:
    def __init__(
        self,
        theater: "Theater",
        session_state: "SessionState",
        animal_state: "AnimalState",
    ) -> None:
        self._theater = theater
        self._session_config = session_state
        self._screen_type = self._session_config.session_config.screen_type

        self._on_trial_start()

    def _on_trial_start(self) -> None:
        self._create_view()

    def _create_view(self) -> None:
        self._background = Canvas(
            self._theater.root,
            bg="red",
            width=self._screen_type.width,
            height=self._screen_type.height,
            highlightthickness=0,
        )
        self._background.place(relx=0.5, rely=0.5, anchor="center")

    def start(self) -> "Feedback":
        self._theater.root.mainloop()

        return True

    def quit(self) -> None:
        self._background.destroy()
        self._theater.root.quit()

    def on_idle(self) -> None:
        self.quit()

    def on_return(self) -> None: ...

    @property
    def condition(self) -> "ScheduleCondition | None":
        return None



from datetime import datetime
from math import ceil
from tkinter import CENTER, Canvas, Event
from typing import TYPE_CHECKING, Final

from mxbi.tasks.GNGSiD.models import Result, TouchEvent
from mxbi.tasks.GNGSiD.tasks.touch.touch_models import DataToShow, TrialData
from mxbi.tasks.GNGSiD.tasks.utils.targets import DetectTarget
from mxbi.utils.aplayer import ToneConfig
from mxbi.utils.tkinter.components.canvas_with_border import CanvasWithInnerBorder
from mxbi.utils.tkinter.components.showdata_widget import ShowDataWidget

if TYPE_CHECKING:
    from concurrent.futures import Future

    from mxbi.models.animal import AnimalState
    from mxbi.models.session import ScreenConfig, SessionConfig
    from mxbi.tasks.GNGSiD.models import PersistentData
    from mxbi.tasks.GNGSiD.tasks.touch.touch_models import TrialConfig
    from mxbi.theater import Theater
    from numpy import int16
    from numpy.typing import NDArray


class GNGSiDTouchScene:
    def __init__(
        self,
        theater: "Theater",
        session: "SessionConfig",
        animal_state: "AnimalState",
        screen_type: "ScreenConfig",
        trial_config: "TrialConfig",
        persistent_data: "PersistentData",
    ) -> None:
        self._theater: "Final[Theater]" = theater
        self._animal_state: "Final[AnimalState]" = animal_state
        self._screen_type: "Final[ScreenConfig]" = screen_type
        self._trial_config: "Final[TrialConfig]" = trial_config
        self._persistent_data: "Final[PersistentData]" = persistent_data

        self._tone = self._prepare_stimulus()
        self._standard_reward_stimulus = self._theater.new_standard_reward_stimulus(
            self._trial_config.stimulus_duration
        )

        self._set_stimulus_intensity()

        self._on_trial_start()

    # region public api
    def start(self) -> TrialData:
        self._theater.mainloop()

        return self._data

    def cancle(self) -> None:
        self._data.result = Result.CANCEL
        self._theater.aplayer.stop()
        self._on_trial_end()

    # endregion

    # region lifecycle
    def _on_trial_start(self) -> None:
        self._create_view()
        self._bind_events()
        self._init_data()

    def _on_inter_trial(self) -> None:
        self._background.after(
            self._trial_config.inter_trial_interval, self._on_trial_end
        )

    def _on_trial_end(self) -> None:
        self._background.destroy()
        self._theater.root.quit()

    # endregion

    # region views
    def _create_view(self) -> None:
        self._create_background()
        self._create_show_data_widget()
        self._create_target()

    def _create_background(self) -> None:
        self._background = CanvasWithInnerBorder(
            master=self._theater.root,
            bg="black",
            width=self._screen_type.width,
            height=self._screen_type.height,
            border_width=40,
        )

        self._background.place(relx=0.5, rely=0.5, anchor="center")

    def _create_show_data_widget(self) -> None:
        self._show_data_widget = ShowDataWidget(self._background)
        self._show_data_widget.place(relx=0, rely=1.2, anchor="sw")
        data = DataToShow(
            name=self._animal_state.name,
            id=self._animal_state.trial_id,
            level_id=self._animal_state.current_level_trial_id,
            level=self._trial_config.level,
            rewards=self._persistent_data.rewards,
            correct=self._animal_state.correct_trial,
            incorrect=self._persistent_data.incorrect,
            timeout=self._persistent_data.timeout,
        )
        self._show_data_widget.show_data(data.model_dump())

    def _create_target(self) -> None:
        xshift = 240
        xcenter = self._screen_type.width * 0.5 + xshift
        ycenter = self._screen_type.height * 0.5

        self._trigger_canvas = DetectTarget(
            self._background, self._trial_config.stimulation_size
        )
        self._trigger_canvas.place(x=xcenter, y=ycenter, anchor="center")

    def _create_wrong_view(self) -> None:
        self._trigger_canvas = Canvas(
            self._background,
            bg="grey",
            width=self._screen_type.width,
            height=self._screen_type.height,
        )
        self._trigger_canvas.place(relx=0.5, rely=0.5, anchor=CENTER)

    # endregion

    # region event binding
    def _bind_events(self) -> None:
        # Manual reward
        self._background.focus_set()
        self._background.bind("<r>", lambda e: self._give_standard_stimulus())

        # Trigger event
        self._background.bind("<ButtonPress>", self._on_background_touched)
        self._trigger_canvas.bind("<ButtonPress>", self._on_touched)

        # Timeout event
        self._trigger_canvas.after(self._trial_config.time_out, self._on_timeout)

    # endregion

    # region event handlers
    def _on_touched(self, event: Event) -> None:
        self._data.touch_events.append(
            TouchEvent(time=datetime.now().timestamp(), x=event.x, y=event.y)
        )
        self._background.unbind("<ButtonPress>")
        self._trigger_canvas.destroy()

        future = self._give_stimulus()
        future.add_done_callback(self._on_stimulus_complete)

    def _on_background_touched(self, event: Event) -> None:
        self._data.touch_events.append(
            TouchEvent(time=datetime.now().timestamp(), x=event.x, y=event.y)
        )
        self._background.unbind("<ButtonPress>")
        self._trigger_canvas.destroy()

        self._on_incorrect()

    def _on_correct(self) -> None:
        self._give_reward()
        self._data.result = Result.CORRECT
        self._data.correct_rate = (self._animal_state.correct_trial + 1) / (
            self._animal_state.current_level_trial_id + 1
        )

        self._on_inter_trial()

    def _on_incorrect(self) -> None:
        self._data.result = Result.INCORRECT
        self._data.correct_rate = self._animal_state.correct_trial / (
            self._animal_state.current_level_trial_id + 1
        )

        self._create_wrong_view()
        self._on_inter_trial()

    def _on_timeout(self) -> None:
        self._data.result = Result.TIMEOUT
        self._data.correct_rate = self._animal_state.correct_trial / (
            self._animal_state.current_level_trial_id + 1
        )

        self._create_wrong_view()
        self._on_inter_trial()

    # endregion

    # region sitimulus and reward
    def _prepare_stimulus(self) -> "NDArray[int16]":
        unit_duration = (
            self._trial_config.stimulus_freq_duration
            + self._trial_config.stimulus_freq_duration
        )

        times = ceil(self._trial_config.stimulus_duration / unit_duration)
        times = max(times, 1)

        freq_1 = ToneConfig(
            frequency=self._trial_config.stimulus_freq,
            duration=self._trial_config.stimulus_freq_duration,
        )
        freq_2 = ToneConfig(frequency=0, duration=self._trial_config.stimulus_interval)

        return self._theater.aplayer.generate_stimulus([freq_1, freq_2], times)

    def _give_stimulus(self) -> "Future[bool]":
        return self._theater.aplayer.play_stimulus(self._tone)

    def _on_stimulus_complete(self, future: "Future[bool]") -> None:
        if future.result():
            self._trigger_canvas.after(
                self._trial_config.reward_delay,
                lambda: self._on_correct(),
            )

    def _give_reward(self) -> None:
        self._persistent_data.rewards += 1
        self._theater.reward.give_reward(self._trial_config.reward_duration)

    def _give_standard_stimulus(self) -> None:
        self._standard_reward_stimulus.play(self._trial_config.reward_duration)

    def _set_stimulus_intensity(self) -> None:
        self._theater.acontroller.set_master_volume(
            self._trial_config.stimulus_freq_master_amp
        )
        self._theater.acontroller.set_digital_volume(
            self._trial_config.stimulus_freq_digital_amp
        )

    # endregion

    # region data
    def _init_data(self) -> None:
        self._data = TrialData(
            animal=self._animal_state.name,
            trial_id=self._animal_state.trial_id,
            current_level_trial_id=self._animal_state.current_level_trial_id,
            trial_config=self._trial_config,
            trial_start_time=datetime.now().timestamp(),
            trial_end_time=0,
            result=Result.TIMEOUT,
            correct_rate=0,
            touch_events=[],
        )

    # endregion



from pathlib import Path
from typing import Dict, List

from mxbi.config import Configure
from mxbi.models.animal import ScheduleCondition
from mxbi.tasks.GNGSiD.models import LevelID, MonkeyName
from pydantic import BaseModel, ConfigDict, RootModel

CONFIG_PATH: Path = Path(__file__).parent / "config.json"


class StimulusConfig(BaseModel):
    stimulus_freq_high: int
    stimulus_freq_high_duration: int
    stimulus_freq_low: int
    stimulus_freq_low_duration: int


class DiscriminateStageParams(BaseModel):
    model_config = ConfigDict(frozen=True)

    # visual stimulation config
    stimulation_size: int
    visual_stimulus_delay: int

    # dynamic reward config
    medium_reward_duration: int
    medium_reward_threshold: int
    low_reward_duration: int

    # audio stimulation config
    min_stimulus_duration: int
    max_stimulus_duration: int
    stimulus_configs: List[StimulusConfig]
    stimulus_interval: int
    extra_response_time: int

    # trial lifecycle
    time_out: int
    inter_trial_interval: int

    # reward config
    reward_duration: int
    reward_delay: int

    attention_duration: int


class DiscriminateStageLeveledParams(BaseModel):
    model_config = ConfigDict(frozen=True)

    level: int

    stimulus_trial_prob: float
    nostimulus_trial_prob: float


class DiscriminateStageConfig(BaseModel):
    model_config = ConfigDict(frozen=True)

    condition: ScheduleCondition

    params: DiscriminateStageParams
    levels_table: Dict[LevelID, DiscriminateStageLeveledParams]


class DiscriminateStageConfigs(RootModel):
    model_config = ConfigDict(frozen=True)

    root: Dict[MonkeyName, DiscriminateStageConfig]


def load_config() -> DiscriminateStageConfigs:
    configs = Configure(CONFIG_PATH, DiscriminateStageConfigs).value
    for config in configs.root.values():
        config.condition.level_count = len(config.levels_table)
    return configs


config = load_config()



from enum import StrEnum, auto
from typing import TypeAlias

from pydantic import BaseModel

LevelID: TypeAlias = int
MonkeyName: TypeAlias = str


class Result(StrEnum):
    CORRECT = auto()
    INCORRECT = auto()
    TIMEOUT = auto()
    CANCEL = auto()


class TouchEvent(BaseModel):
    time: float
    x: int
    y: int


class BaseTrialConfig(BaseModel):
    # basic config
    level: int = 0

    # visual stimulation config
    stimulation_size: int = 0

    # audio stimulation config
    stimulus_duration: int = 0

    # trial lifecycle
    time_out: int = 0
    inter_trial_interval: int = 0

    # reward config
    reward_duration: int = 0
    reward_delay: int = 0


class BaseTrialData(BaseModel):
    animal: str
    trial_id: int
    current_level_trial_id: int
    trial_start_time: float
    trial_end_time: float
    result: Result
    correct_rate: float
    touch_events: list[TouchEvent]


class BaseDataToShow(BaseModel):
    name: str
    id: int
    level_id: int
    level: int
    rewards: int
    correct: int
    incorrect: int
    timeout: int


class PersistentData(BaseModel):
    rewards: int
    correct: int
    incorrect: int
    timeout: int



from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from mxbi.models.animal import AnimalState, ScheduleCondition
    from mxbi.models.session import SessionState
    from mxbi.models.task import Feedback
    from mxbi.theater import Theater


class Task(Protocol):
    def __init__(
        self,
        theater: "Theater",
        session_state: "SessionState",
        animal_state: "AnimalState",
    ) -> None: ...
    def start(self) -> "Feedback": ...

    def quit(self) -> None: ...

    def on_idle(self) -> None: ...

    def on_return(self) -> None: ...

    @property
    def condition(self) -> "ScheduleCondition | None": ...



from pymxbi import MXBIModel
from pymxbi.detector import DetectorEnum, DetectorModel
from pymxbi.rewarder import RewarderEnum, RewarderModel
from PySide6.QtCore import Signal
from PySide6.QtWidgets import (
    QHBoxLayout,
    QMainWindow,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from ..core.config_store import ConfigStore
from ..core.path import get_mxbi_config_path, get_options_session_path
from ..models.session import Options
from .components.baseconfig import BaseConfig
from .components.device_card import (
    BeambreakDetectorCard,
    FusionDetectorCard,
    MockDetectorCard,
    MockRewarderCard,
    RFIDDetectorCard,
    RPIGpioPumpCard,
)
from .components.devices import Devices


class MXBIPanel(QMainWindow):
    accepted = Signal()
    rejected = Signal()

    _REWARDER_CARD_FACTORIES: dict[str, type[QWidget]] = {
        RewarderEnum.RPI_GPIO: RPIGpioPumpCard,
        RewarderEnum.MOCK: MockRewarderCard,
    }
    _DETECTOR_CARD_FACTORIES: dict[str, type[QWidget]] = {
        DetectorEnum.RFID_CONTINUOUS: RFIDDetectorCard,
        DetectorEnum.BEAMBREAK_CONTINUOUS: BeambreakDetectorCard,
        DetectorEnum.FUSION_CONTINUOUS: FusionDetectorCard,
        DetectorEnum.MOCK: MockDetectorCard,
    }

    # -----------------------------
    # Lifecycle / Init
    # -----------------------------

    def __init__(self):
        super().__init__()
        self._accepted = False
        self._config = ConfigStore(get_mxbi_config_path(), MXBIModel)
        self._options = ConfigStore(get_options_session_path(), Options)

        self._build_ui()
        self._load_from_config()
        self._bind_events()

    # -----------------------------
    # UI Construction
    # -----------------------------

    def _build_ui(self) -> None:
        self.setWindowTitle("MXBI Configuration Panel")

        self._widget_main = QWidget()
        self._layout_main = QVBoxLayout()
        self._widget_main.setLayout(self._layout_main)
        self.setCentralWidget(self._widget_main)

        self.base_config = BaseConfig(self, self._options.value.mxbis)
        self._layout_main.addWidget(self.base_config)
        self._build_device_groups()
        self._layout_main.addLayout(self._build_buttons_row())

    def _build_device_groups(self) -> None:
        self.rewarders_group = Devices[RewarderModel](
            self,
            "Rewarders",
            action_label="Add Rewarder",
            device_types=list(RewarderEnum),
            dialog_title="Add rewarder",
            label="rewarder type:",
            card_factories=self._REWARDER_CARD_FACTORIES,
            columns=4,
        )
        self._layout_main.addWidget(self.rewarders_group)

        self.detectors_group = Devices[DetectorModel](
            self,
            "Detectors",
            action_label="Add Detector",
            device_types=list(DetectorEnum),
            dialog_title="Add detector",
            label="detector type:",
            card_factories=self._DETECTOR_CARD_FACTORIES,
            columns=4,
        )
        self._layout_main.addWidget(self.detectors_group)

    def _build_buttons_row(self) -> QHBoxLayout:
        layout_buttons = QHBoxLayout()
        self.save_button = QPushButton("Save")
        self.cancel_button = QPushButton("Cancel")
        self.continue_button = QPushButton("Continue")
        layout_buttons.addWidget(self.cancel_button)
        layout_buttons.addWidget(self.save_button)
        layout_buttons.addWidget(self.continue_button)
        return layout_buttons

    # -----------------------------
    # Load / Save
    # -----------------------------

    def _load_from_config(self) -> None:
        self.base_config.load_from_model(self._config.value)

        self.rewarders_group.load_models(self._config.value.rewarders)
        self.detectors_group.load_models(self._config.value.detectors)

    def _collect_result(self) -> None:
        self.base_config.apply_to_model(self._config.value)

        # Replace instead of append so repeated Save/Continue doesn't duplicate entries.
        self._config.value.rewarders = self.rewarders_group.results()
        self._config.value.detectors = self.detectors_group.results()

    def _on_save(self) -> None:
        self._collect_result()
        self._config.save()

    def _on_continue(self) -> None:
        self._collect_result()
        self._config.save()
        self._accepted = True
        self.close()
        self.accepted.emit()

    # -----------------------------
    # Events / Menus
    # -----------------------------

    def _bind_events(self) -> None:
        self.cancel_button.clicked.connect(self._on_cancel)
        self.save_button.clicked.connect(self._on_save)
        self.continue_button.clicked.connect(self._on_continue)

    def _on_cancel(self) -> None:
        self.close()

    def closeEvent(self, event) -> None:
        super().closeEvent(event)
        if event.isAccepted() and not self._accepted:
            self.rejected.emit()



from __future__ import annotations

from pymxbi import MXBIModel
from pymxbi.screen import get_screen_size
from PySide6.QtCore import Signal
from PySide6.QtWidgets import QComboBox, QFormLayout, QGroupBox, QLabel


class BaseConfig(QGroupBox):
    changed = Signal(str)

    def __init__(self, parent, mxbi_options: list[str]):
        super().__init__(parent)

        self.setTitle("Base config")

        self._layout = QFormLayout()
        self.setLayout(self._layout)

        self._label_mxbi = QLabel("mxbi id:")
        self._combo_mxbi = QComboBox()
        self._combo_mxbi.addItems(mxbi_options)
        self._layout.addRow(self._label_mxbi, self._combo_mxbi)

        self._label_screen = QLabel("screen:")
        self._combo_screen = QComboBox()

        for screen in get_screen_size():
            self._combo_screen.addItem(
                f"{screen.width} * {screen.height}", (screen.width, screen.height)
            )
        self._layout.addRow(self._label_screen, self._combo_screen)

        self._bind_events()

    def _emit_changed(self, msg: str) -> None:
        self.changed.emit(msg)

    def _bind_events(self) -> None:
        self._combo_mxbi.currentTextChanged.connect(self._emit_changed)
        self._combo_screen.currentTextChanged.connect(self._emit_changed)

    @property
    def mxbi_id(self) -> str:
        return self._combo_mxbi.currentText()

    @property
    def screen_size(self) -> tuple[int, int]:
        return self._combo_screen.currentData()

    def load_from_model(self, model: MXBIModel) -> None:
        self._combo_mxbi.setCurrentText(str(model.mxbi_id))
        for i in range(self._combo_screen.count()):
            if self._combo_screen.itemData(i) == tuple(model.screen_size):
                self._combo_screen.setCurrentIndex(i)
                break

    def apply_to_model(self, model: MXBIModel) -> None:
        mxbi_id_text = self._combo_mxbi.currentText().strip()
        try:
            model.mxbi_id = int(mxbi_id_text) if mxbi_id_text else 0
        except ValueError:
            model.mxbi_id = 0

        model.screen_size = self._combo_screen.currentData()



from pygame import Event, Surface

from .idle.idle import IDLE
from .scene_protocol import SceneProtocol


class SceneManager:
    _scenes: dict[str, type[SceneProtocol]] = {}

    def __init__(self) -> None:
        self.current: SceneProtocol | None = None
        self._pending: SceneProtocol | None = None
        self._default_scene: type[SceneProtocol] = IDLE
        self._unknown_animal_fallback: type[SceneProtocol] = IDLE
        self._fault_fallback: type[SceneProtocol] = IDLE

        self.register([IDLE])

    def init(self) -> None:
        self.switch(self._default_scene, defer=False)

    def register(
        self, scene: dict[str, type[SceneProtocol]] | list[type[SceneProtocol]]
    ) -> None:
        if isinstance(scene, dict):
            self._scenes.update(scene)
        elif isinstance(scene, list):
            for s in scene:
                self._scenes[s.__name__.lower()] = s
        else:
            raise ValueError("Invalid scene type")

    @property
    def scenes(self) -> dict[str, type[SceneProtocol]]:
        return self._scenes

    @property
    def stage_level_tables(self) -> dict[str, dict[str, list[int]]]:
        return {name: scene.level_table for name, scene in self._scenes.items()}

    @property
    def default_scene(self) -> type[SceneProtocol]:
        return self._default_scene

    @default_scene.setter
    def default_scene(self, name: str) -> None:
        self._default_scene = self._scenes[name]

    @property
    def unknown_animal_fallback(self) -> type[SceneProtocol]:
        return self._unknown_animal_fallback

    @unknown_animal_fallback.setter
    def unknown_animal_fallback(self, name: str) -> None:
        self._unknown_animal_fallback = self._scenes[name]

    @property
    def fault_fallback(self) -> type[SceneProtocol]:
        return self._fault_fallback

    @fault_fallback.setter
    def fault_fallback(self, name: str) -> None:
        self._fault_fallback = self._scenes[name]

    def switch(self, scene: type[SceneProtocol], defer: bool = True) -> None:
        if defer:
            self._pending = scene()
        else:
            self._switch(scene())

    def _switch(self, next_scene: SceneProtocol) -> None:
        prev = self.current
        if prev is not None and prev.running:
            prev.quit()

        self.current = next_scene
        next_scene.start()

    def apply_pending(self) -> None:
        if self._pending is None:
            return
        next_scene = self._pending
        self._pending = None
        self._switch(next_scene)

    def handle_event(self, event: Event) -> None:
        if self.current:
            self.current.handle_event(event)

    def update(self, dt_s: float) -> None:
        if self.current:
            self.current.update(dt_s)

    def draw(self, screen: Surface) -> None:
        if self.current:
            self.current.draw(screen)



from pymxbi.detector import BeamBreakContinuousDetectorModel
from PySide6.QtWidgets import QLabel, QLineEdit

from ..device_card import DeviceCard


class BeambreakDetectorCard(DeviceCard[BeamBreakContinuousDetectorModel]):
    def __init__(self):
        super().__init__()

        self.set_title("Beambreak Detector")

        label_pin = QLabel("Pin")
        self._line_pin = QLineEdit()
        self._line_pin.setPlaceholderText("Pin")
        self.layout_config.addRow(label_pin, self._line_pin)

        label_animal_id = QLabel("Animal ID")
        self._line_animal_id = QLineEdit()
        self._line_animal_id.setPlaceholderText("Animal ID")
        self.layout_config.addRow(label_animal_id, self._line_animal_id)

        label_poll_interval = QLabel("Poll Interval")
        self._line_poll_interval = QLineEdit()
        self._line_poll_interval.setPlaceholderText("Poll Interval (s)")
        self.layout_config.addRow(label_poll_interval, self._line_poll_interval)

    def load_config(self, model: BeamBreakContinuousDetectorModel) -> None:
        self.checkbox_enabled.setChecked(model.enabled)
        self.line_device_id.setText(str(model.id))
        self._line_pin.setText(str(model.pin))
        self._line_animal_id.setText(model.animal_id)
        self._line_poll_interval.setText(str(model.poll_interval))

    @property
    def result(self) -> BeamBreakContinuousDetectorModel:
        return BeamBreakContinuousDetectorModel(
            enabled=self.checkbox_enabled.isChecked(),
            id=int(self.line_device_id.text()),
            pin=int(self._line_pin.text()),
            animal_id=self._line_animal_id.text(),
            poll_interval=float(self._line_poll_interval.text()),
        )



from datetime import datetime
from random import choices
from typing import TYPE_CHECKING, Final

from mxbi.data_logger import DataLogger, DataLoggerType
from mxbi.models.animal import ScheduleCondition
from mxbi.tasks.default.initial_habituation_training.stages.models import (
    InitialHabituationTrainingStageConfig,
    StageContext,
    StageContexts,
    config,
)
from mxbi.tasks.default.initial_habituation_training.tasks.stay_to_reward.stay_to_reward import (
    DefaultStayToRewardScene,
)
from mxbi.tasks.default.initial_habituation_training.tasks.stay_to_reward.stay_to_reward_models import (
    Result,
    TrialConfig,
)
from mxbi.utils.logger import logger
from mxbi.utils.tkinter.components.canvas_with_border import CanvasWithInnerBorder

if TYPE_CHECKING:
    from mxbi.models.animal import AnimalState
    from mxbi.models.session import SessionConfig, SessionState
    from mxbi.models.task import Feedback
    from mxbi.theater import Theater


contexts = StageContexts()
background: CanvasWithInnerBorder | None = None


def _initialize_contexts(session_config: "SessionConfig") -> None:
    global contexts
    for monkey in session_config.animals.keys():
        if contexts.root.get(monkey) is None:
            contexts.root[monkey] = StageContext()


def _initialize_background(theater: "Theater", session_config: "SessionConfig") -> None:
    global background
    background = CanvasWithInnerBorder(
        master=theater.root,
        bg="black",
        width=session_config.screen_type.width,
        height=session_config.screen_type.height,
        border_width=40,
    )


class InitialHabituationTrainingStage:
    STAGE_NAME: Final[str] = "DEFAULT_INITIAL_HABITUATION_TRAINING_STAGE"

    def __init__(
        self,
        theater: "Theater",
        session_state: "SessionState",
        animal_state: "AnimalState",
    ) -> None:
        self._theater = theater
        self._session_state = session_state
        self._animal_state = animal_state

        if not contexts.root:
            _initialize_contexts(session_state.session_config)

        if background is None:
            _initialize_background(theater, session_state.session_config)

        context = contexts.root[animal_state.name]

        self._stage_config = self._load_stage_config(self._animal_state.name)

        _levels_config = self._stage_config.levels_table[animal_state.level]
        self._stage_config.condition.config.evaluation_interval = (
            _levels_config.evaluation_interval
        )

        entry_reward = choices(
            [True, False],
            weights=[_levels_config.entry_reward, 1 - _levels_config.entry_reward],
        )[0]

        _config = TrialConfig(
            level=_levels_config.level,
            entry_reward=entry_reward,
            min_stimulus_interval=_levels_config.min_stimulus_interval,
            max_stimulus_interval=_levels_config.max_stimulus_interval,
            target=_levels_config.target,
            stimulus_density=_levels_config.stimulus_density,
        )

        self._data_logger = DataLogger(
            self._session_state,
            self._animal_state.name,
            self.STAGE_NAME,
            DataLoggerType.JSONL,
        )

        if self._animal_state.data_path is None:
            self._animal_state.data_path = self._data_logger.path

        assert background is not None

        self._task = DefaultStayToRewardScene(
            theater=theater,
            animal_state=animal_state,
            screen_type=session_state.session_config.screen_type,
            trial_config=_config,
            context=context,
            background=background,
        )

    def start(self) -> "Feedback":
        trial_data = self._task.start()
        self._data_logger.save(trial_data.model_dump())

        feedback = self._handle_result(trial_data.result)
        logger.debug(
            f"{self.STAGE_NAME}: "
            f"session_id={self._session_state.session_id}, "
            f"animal_name={self._animal_state.name}, "
            f"animal_level={self._animal_state.level}, "
            f"state_name={self.STAGE_NAME}, "
            f"result={trial_data}, "
            f"feedback={feedback}"
        )

        now = datetime.now().timestamp()
        if self._animal_state.animal_session_start_time != 0.0:
            session_duration = now - self._animal_state.animal_session_start_time
            if session_duration >= 30:
                print(session_duration)
                print(self._animal_state.animal_session_start_time)
                self._animal_state.animal_session_start_time = (
                    datetime.now().timestamp()
                )

                self._theater._scheduler._increase_difficulty(self._animal_state)

        return feedback

    def _load_stage_config(self, monkey: str) -> InitialHabituationTrainingStageConfig:
        stage_config = config.root.get(monkey) or config.root.get("default")
        if stage_config is None:
            raise ValueError("No default stage config found")
        return stage_config

    def _handle_result(self, result: "Result") -> "Feedback":
        feedback = False

        match result:
            case Result.CORRECT:
                feedback = True
            case Result.INCORRECT:
                feedback = False

        return feedback

    def quit(self) -> None:
        self._task.cancle()

    def on_idle(self) -> None:
        self._task.cancle()

    def on_return(self) -> None:
        self._task.cancle()

    @property
    def condition(self) -> "ScheduleCondition | None":
        return self._stage_config.condition



from pymxbi.detector import RFIDContinuousDetectorModel
from PySide6.QtWidgets import (
    QComboBox,
    QLabel,
    QLineEdit,
)

from .....utils.serial import get_all_ports, get_baudrates
from ..device_card import DeviceCard


class RFIDDetectorCard(DeviceCard[RFIDContinuousDetectorModel]):
    def __init__(self):
        super().__init__()
        self.set_title("Single RFID Detector")

        lable_port = QLabel("Port:")
        self.combo_port = QComboBox()
        self.combo_port.addItems(get_all_ports())
        self.layout_config.addRow(lable_port, self.combo_port)

        lable_baudrate = QLabel("Baudrate:")
        self.combo_baudrate = QComboBox()
        for text, value in get_baudrates():
            self.combo_baudrate.addItem(text, value)
        self.layout_config.addRow(lable_baudrate, self.combo_baudrate)

        label_poll_interval = QLabel("Poll Interval:")
        self._line_poll_interval = QLineEdit()
        self._line_poll_interval.setPlaceholderText("Poll Interval (s)")
        self.layout_config.addRow(label_poll_interval, self._line_poll_interval)

        label_max_tag_age = QLabel("Max Tag Age:")
        self._line_max_tag_age = QLineEdit()
        self._line_max_tag_age.setPlaceholderText("Max Tag Age (s)")
        self.layout_config.addRow(label_max_tag_age, self._line_max_tag_age)

    def load_config(self, model: RFIDContinuousDetectorModel) -> None:
        self.checkbox_enabled.setChecked(model.enabled)
        self.line_device_id.setText(str(model.id))
        self.combo_port.setCurrentText(model.port)
        self.combo_baudrate.setCurrentText(str(model.baudrate))
        self._line_poll_interval.setText(str(model.poll_interval))
        self._line_max_tag_age.setText(str(model.max_tag_age_seconds))

    @property
    def result(self) -> RFIDContinuousDetectorModel:
        return RFIDContinuousDetectorModel(
            enabled=self.checkbox_enabled.isChecked(),
            id=int(self.line_device_id.text()),
            port=self.combo_port.currentText(),
            baudrate=int(self.combo_baudrate.currentText()),
            poll_interval=float(self._line_poll_interval.text()),
            max_tag_age_seconds=float(self._line_max_tag_age.text()),
        )



import time
from pathlib import Path
from typing import TYPE_CHECKING, Final

from mxbi.data_logger import DataLogger
from mxbi.tasks.cross_modal.bundle_dir import CrossModalBundleDir
from mxbi.tasks.cross_modal.config import CrossModalConfig, load_cross_modal_config
from mxbi.tasks.cross_modal.media import load_wav_as_int16
from mxbi.tasks.cross_modal.models import CrossModalOutcome, CrossModalResultRecord
from mxbi.tasks.cross_modal.scene import CrossModalResult, CrossModalScene
from mxbi.tasks.cross_modal.trial_io import TrialCursor
from mxbi.utils.logger import logger
from PIL import Image

if TYPE_CHECKING:
    import numpy as np
    from mxbi.models.animal import AnimalState
    from mxbi.models.session import SessionState
    from mxbi.models.task import Feedback
    from mxbi.theater import Theater
    from numpy.typing import NDArray
    from PIL.Image import Image as PILImage


class CrossModalTask:
    STAGE_NAME: Final[str] = "cross_modal_task"

    def __init__(
        self,
        theater: "Theater",
        session_state: "SessionState",
        animal_state: "AnimalState",
    ) -> None:
        self._theater = theater
        self._session_state = session_state
        self._animal_state = animal_state
        self._screen = session_state.session_config.screen_type

        bundle_dir_str = getattr(
            session_state.session_config, "cross_modal_bundle_dir", None
        )
        if not bundle_dir_str:
            raise RuntimeError(
                "No cross-modal bundle directory configured. "
                "Set session_config.cross_modal_bundle_dir via the launcher."
            )

        bundle_root = Path(bundle_dir_str).expanduser().resolve()
        self._bundle_dir = CrossModalBundleDir.from_dir_path(bundle_root)

        subject_id = self._animal_state.name
        trials = self._bundle_dir.load_trials(subject_id)
        if not trials:
            raise RuntimeError(
                f"Bundle contains zero trials for subject '{subject_id}'."
            )

        self._cursor = TrialCursor(
            bundle_root=self._bundle_dir.root_dir, subject_id=subject_id
        )
        self._trial_index = self._cursor.next_index(len(trials))
        self._trial = trials[self._trial_index]

        self._cross_modal_config: CrossModalConfig = load_cross_modal_config()

        image_size = int(
            min(self._screen.width, self._screen.height)
            * self._cross_modal_config.visual.image_scale
        )

        left_image_path = self._bundle_dir.resolve_media_path(
            self._trial.left_image_path
        )
        right_image_path = self._bundle_dir.resolve_media_path(
            self._trial.right_image_path
        )
        audio_path = self._bundle_dir.resolve_media_path(self._trial.audio_path)

        left_image = self._prepare_image(
            left_image_path,
            image_size=image_size,
        )
        right_image = self._prepare_image(
            right_image_path,
            image_size=image_size,
        )

        audio_stimulus = load_wav_as_int16(
            audio_path,
            rate_policy=self._cross_modal_config.audio.wav_rate_policy,
            gain=self._cross_modal_config.audio.gain,
        )

        self._scene = CrossModalScene(
            theater=self._theater,
            session_state=self._session_state,
            animal_state=self._animal_state,
            screen=self._screen,
            trial=self._trial,
            cross_modal_config=self._cross_modal_config,
            left_image=left_image,
            right_image=right_image,
            audio_stimulus=audio_stimulus,
        )

        self._data_logger = DataLogger(
            self._session_state, self._animal_state.name, self.STAGE_NAME
        )
        self._feedback = False

    def start(self) -> "Feedback":
        result = self._scene.start()
        self._feedback = result.feedback

        if not result.cancelled:
            self._log_trial(result)

        self._cursor.advance(self._trial_index)

        logger.debug(
            "cross_modal_task: session_id=%s, subject=%s, level=%s, "
            "trial_index=%s, is_partner=%s, feedback=%s",
            getattr(self._session_state, "session_id", None),
            self._animal_state.name,
            self._animal_state.level,
            self._trial_index + 1,
            self._trial.is_partner_call,
            self._feedback,
        )
        return self._feedback

    def quit(self) -> None:
        self._scene.cancel()

    def on_idle(self) -> None:
        self._scene.cancel()

    def on_return(self) -> None:
        self._scene.cancel()

    @property
    def condition(self):
        return None

    def _log_trial(self, result: CrossModalResult) -> None:
        try:
            if result.timeout:
                outcome = CrossModalOutcome.TIMEOUT
            elif result.cancelled:
                outcome = CrossModalOutcome.ABORTED
            elif result.feedback:
                outcome = CrossModalOutcome.CORRECT
            else:
                outcome = CrossModalOutcome.INCORRECT

            if (
                result.trial_start_time is not None
                and result.choice_time is not None
                and not result.timeout
            ):
                latency = result.choice_time - result.trial_start_time
            else:
                latency = None

            if result.chosen_side == "left":
                chosen_identity = self._trial.left_image_identity_id
            elif result.chosen_side == "right":
                chosen_identity = self._trial.right_image_identity_id
            else:
                chosen_identity = None

            rec = CrossModalResultRecord(
                timestamp=time.time(),
                session_id=getattr(self._session_state, "session_id", None),
                subject_id=self._trial.subject_id,
                partner_id=self._trial.partner_id,
                trial_id=self._trial.trial_id,
                trial_number=self._trial.trial_number,
                call_identity_id=self._trial.call_identity_id,
                call_category=self._trial.call_category,
                is_partner_call=self._trial.is_partner_call,
                other_identity_id=self._trial.other_identity_id,
                other_category=self._trial.other_category,
                partner_side=self._trial.partner_side,
                correct_side=self._trial.correct_side,
                audio_identity_id=self._trial.audio_identity_id,
                audio_index=self._trial.audio_index,
                audio_path=self._trial.audio_path,
                left_image_identity_id=self._trial.left_image_identity_id,
                left_image_index=self._trial.left_image_index,
                left_image_path=self._trial.left_image_path,
                right_image_identity_id=self._trial.right_image_identity_id,
                right_image_index=self._trial.right_image_index,
                right_image_path=self._trial.right_image_path,
                chosen_side=result.chosen_side,
                chosen_identity_id=chosen_identity,
                outcome=outcome,
                trial_start_time=result.trial_start_time,
                choice_time=result.choice_time,
                latency_sec=latency,
                choice_x=result.choice_x,
                choice_y=result.choice_y,
                aborted=result.cancelled,
                timeout=result.timeout,
            )

            payload = rec.model_dump()
            self._data_logger.save_jsonl(payload)
            self._data_logger.save_csv_row(payload)
        except Exception:
            logger.exception("Failed to log cross-modal trial")

    @staticmethod
    def _prepare_image(image_path: Path, *, image_size: int) -> "PILImage":
        with Image.open(image_path) as img:
            prepared = img.convert("RGB").rotate(90, expand=True)
            prepared = prepared.resize((image_size, image_size), Image.LANCZOS)
            return prepared



from typing import Protocol, runtime_checkable

from pygame.event import Event
from pygame.surface import Surface


@runtime_checkable
class SceneProtocol(Protocol):
    _running: bool
    level_table: dict[str, list[int]]

    def start(self) -> None: ...

    def quit(self) -> None: ...

    @property
    def running(self) -> bool: ...

    def handle_event(self, event: Event) -> None: ...

    def update(self, dt_s: float) -> None: ...

    def draw(self, screen: Surface) -> None: ...



from pathlib import Path

from mxbi.config import Configure
from mxbi.models.animal import ScheduleCondition
from mxbi.tasks.GNGSiD.models import LevelID, MonkeyName
from pydantic import BaseModel, ConfigDict, RootModel

CONFIG_PATH = Path(__file__).parent / "config.json"


class DetectStageParams(BaseModel):
    model_config = ConfigDict(frozen=True)

    # visual stimulation config
    stimulation_size: int
    visual_stimulus_delay: int

    # audio stimulation config
    stimulus_freq: int
    stimulus_freq_duration: int
    stimulus_interval: int

    # trial lifecycle
    time_out: int
    inter_trial_interval: int

    # reward config
    reward_duration: int
    reward_delay: int


class DetectStageLeveledParams(BaseModel):
    model_config = ConfigDict(frozen=True)

    level: int
    min_stimulus_duration: int
    max_stimulus_duration: int

    go_task_prob: float
    nogo_task_prob: float


class DetectStageConfig(BaseModel):
    model_config = ConfigDict(frozen=True)

    condition: ScheduleCondition

    params: DetectStageParams
    levels_table: dict[LevelID, DetectStageLeveledParams]


class DetectStageConfigs(RootModel):
    model_config = ConfigDict(frozen=True)

    root: dict[MonkeyName, DetectStageConfig]


def load_config() -> DetectStageConfigs:
    configs = Configure(CONFIG_PATH, DetectStageConfigs).value
    for config in configs.root.values():
        config.condition.level_count = len(config.levels_table)
    return configs


config = load_config()



import os
from pathlib import Path
from typing import Generic, TypeVar

from pydantic import BaseModel

T = TypeVar("T", bound=BaseModel)


class ConfigStore(Generic[T]):
    def __init__(
        self,
        config_path: Path,
        config_class: type[T],
        *,
        create_default: bool = True,
    ) -> None:
        self._config_path = config_path
        self._config_class = config_class
        self._create_default = create_default
        self._config = self._load_config()

    @property
    def value(self) -> T:
        return self._config

    def _ensure_config_readable(self) -> None:
        if not self._config_path.exists():
            raise FileNotFoundError(f"Config file {self._config_path} not found")

        if not os.access(self._config_path, os.R_OK):
            raise PermissionError(f"Config file {self._config_path} is not readable")

    def _create_default_config(self) -> T:
        config = self._config_class()
        self._config_path.parent.mkdir(parents=True, exist_ok=True)
        self._config_path.write_text(
            config.model_dump_json(indent=4),
            encoding="utf-8",
        )
        return config

    def _load_config(self) -> T:
        try:
            self._ensure_config_readable()
            text = self._config_path.read_text(encoding="utf-8")
            return self._config_class.model_validate_json(text)
        except Exception as e:
            if self._create_default:
                return self._create_default_config()
            raise RuntimeError(
                f"Failed to load config file {self._config_path}: {e}"
            ) from e

    def save(self, data: T | None = None) -> None:
        if data is not None:
            self._config = data

        self._config_path.parent.mkdir(parents=True, exist_ok=True)
        try:
            self._config_path.write_text(
                self._config.model_dump_json(indent=4),
                encoding="utf-8",
            )
        except Exception as e:
            raise RuntimeError(
                f"Failed to save config file {self._config_path}: {e}"
            ) from e



import wave
from pathlib import Path
from typing import Literal

import numpy as np
from mxbi.utils.aplayer import SAMPLE_RATE
from numpy.typing import NDArray


def load_wav_as_int16(
    path: Path,
    *,
    target_rate: int = SAMPLE_RATE,
    rate_policy: Literal["resample", "error"] = "resample",
    gain: float = 1.0,
) -> NDArray[np.int16]:
    with wave.open(str(path), "rb") as wf:
        frames = wf.readframes(wf.getnframes())
        sample_width = wf.getsampwidth()
        channel_count = wf.getnchannels()
        source_rate = wf.getframerate()

    if sample_width != 2:
        raise ValueError(f"Only 16-bit WAVs supported, got sampwidth={sample_width}")

    data = np.frombuffer(frames, dtype=np.int16)

    if channel_count > 1:
        data = data.reshape(-1, channel_count).mean(axis=1).astype(np.int16)

    float_data = (data.astype(np.float32)) / float(np.iinfo(np.int16).max)

    if source_rate != target_rate:
        if rate_policy == "error":
            raise ValueError(
                f"WAV sample rate mismatch: wav={source_rate}Hz, expected={target_rate}Hz"
            )
        float_data = _resample_1d(
            float_data, source_rate=source_rate, target_rate=target_rate
        )

    if gain != 1.0:
        float_data = np.clip(float_data * gain, -1.0, 1.0)

    return (float_data * float(np.iinfo(np.int16).max)).astype(np.int16)


def _resample_1d(
    signal: NDArray[np.float32], *, source_rate: int, target_rate: int
) -> NDArray[np.float32]:
    if signal.size == 0:
        return signal

    target_length = int(round(signal.size * (target_rate / source_rate)))
    if target_length <= 0:
        return np.zeros(0, dtype=np.float32)

    old_positions = np.linspace(0.0, 1.0, num=signal.size, endpoint=False)
    new_positions = np.linspace(0.0, 1.0, num=target_length, endpoint=False)
    resampled = np.interp(new_positions, old_positions, signal).astype(np.float32)
    return resampled



from .config_store import ConfigStore
from .context import MXBIFlow, get_mxbiflow, set_mxbiflow
from .path import get_base_path, set_base_path

__all__ = [
    "ConfigStore",
    "MXBIFlow",
    "get_mxbiflow",
    "set_mxbiflow",
    "get_base_path",
    "set_base_path",
]



from datetime import datetime
from math import ceil
from random import uniform
from tkinter import Frame
from typing import TYPE_CHECKING, Final

from mxbi.tasks.default.initial_habituation_training.tasks.stay_to_reward.stay_to_reward_models import (
    DataToShow,
    Result,
    TrialData,
)
from mxbi.utils.aplayer import ToneConfig
from mxbi.utils.tkinter.components.canvas_with_border import CanvasWithInnerBorder
from mxbi.utils.tkinter.components.showdata_widget import ShowDataWidget

if TYPE_CHECKING:
    from concurrent.futures import Future

    from mxbi.models.animal import AnimalState
    from mxbi.models.session import ScreenConfig
    from mxbi.tasks.default.initial_habituation_training.stages.models import (
        StageContext,
    )
    from mxbi.tasks.default.initial_habituation_training.tasks.stay_to_reward.stay_to_reward_models import (
        TrialConfig,
    )
    from mxbi.theater import Theater
    from numpy import int16
    from numpy.typing import NDArray


class DefaultStayToRewardScene:
    def __init__(
        self,
        theater: "Theater",
        animal_state: "AnimalState",
        screen_type: "ScreenConfig",
        trial_config: "TrialConfig",
        context: "StageContext",
        background: "CanvasWithInnerBorder",
    ):
        self._theater: "Final[Theater]" = theater
        self._animal_state: "Final[AnimalState]" = animal_state
        self._screen_type: "Final[ScreenConfig]" = screen_type
        self._trial_config: "Final[TrialConfig]" = trial_config
        self._context: "Final[StageContext]" = context
        self._background = background

        self._play_future: Future[bool] | None = None

        self._tone = self._prepare_stimulus()
        self._standard_reward_stimulus = self._theater.new_standard_reward_stimulus(
            self._trial_config.stimulus_duration
        )

        self._set_stimulus_intensity()
        self._on_trial_start()

    # region public api
    def start(self) -> "TrialData":
        self._theater.mainloop()

        return self._data

    def cancle(self) -> None:
        self._cleanup()

    # endregion

    # region lifecycle
    def _on_trial_start(self) -> None:
        self._create_view()
        self._init_data()
        self._bind_events()
        self._start_tracking_data()

        if (
            self._trial_config.entry_reward
            and self._animal_state.current_animal_session_trial_id == 1
        ):
            self._direct_stimulus()
        else:
            self._start_timing()
            self._stimulus_loop()

    def _on_trial_end(self) -> None:
        if self._play_future is not None:
            self._play_future.add_done_callback(lambda _: self._cleanup())
        else:
            self._cleanup()

    def _cleanup(self) -> None:
        self._data.trial_end_time = datetime.now().timestamp()
        self._data.stay_duration = (
            self._data.trial_end_time - self._data.trial_start_time
        )

        self._theater.aplayer.stop()
        self._trigger.destroy()
        self._show_data_widget.destroy()
        self._theater.root.quit()

    # endregion

    # region views

    def _create_view(self) -> None:
        self._create_background()
        self._create_trigger()
        self._create_show_data_widget()

    def _create_background(self) -> None:
        self._background.place(relx=0.5, rely=0.5, anchor="center")

    def _create_trigger(self) -> None:
        self._trigger = Frame(bg="red", width=0, height=0)
        self._trigger.place(relx=0.5, rely=0.5, anchor="center")
        self._trigger.lower()

    def _create_show_data_widget(self) -> None:
        self._show_data_widget = ShowDataWidget(self._background)
        self._show_data_widget.place(relx=0, rely=1, anchor="sw")
        _data = DataToShow(
            name=self._animal_state.name,
            ald=self._animal_state.current_animal_session_trial_id,
            level=self._animal_state.level,
            id=self._animal_state.trial_id,
            lid=self._animal_state.current_level_trial_id,
            t_dur=f"{self._context.duration} s",
            dur="0 s",
            rewards=0,
        )
        self._show_data_widget.show_data(_data.model_dump())

    # endregion

    # region event binding
    def _bind_events(self) -> None:
        self._trigger.focus_set()

    def _start_timing(self) -> None:
        target_ms = int(self._trial_config.target * 1000)
        self._trigger.after(target_ms, self._on_correct)

    def _stimulus_loop(self) -> None:
        stimulus_interval = uniform(
            self._trial_config.min_stimulus_interval,
            self._trial_config.max_stimulus_interval,
        )
        stimulus_interval_ms = int(stimulus_interval * 1000)

        self._trigger.after(stimulus_interval_ms, self._give_stimulus)

    # endregion

    # region data
    def _start_tracking_data(self) -> None:
        data = DataToShow(
            name=self._animal_state.name,
            ald=self._animal_state.current_animal_session_trial_id,
            level=self._animal_state.level,
            id=self._animal_state.trial_id,
            lid=self._animal_state.current_level_trial_id,
            t_dur=f"{self._context.duration} s",
            dur=f"{int(self._data.stay_duration)} s",
            rewards=self._context.rewards,
        )

        self._data.stay_duration += 1
        self._context.duration += 1

        self._show_data_widget.update_data(data.model_dump())
        self._trigger.after(1000, self._start_tracking_data)

    def _init_data(self) -> None:
        self._data = TrialData(
            level=self._trial_config.level,
            level_trial_id=self._animal_state.current_level_trial_id,
            animal_session_trial_id=self._animal_state.current_animal_session_trial_id,
            animal=self._animal_state.name,
            trial_id=self._animal_state.trial_id,
            trial_start_time=datetime.now().timestamp(),
            trial_end_time=0,
            stay_duration=0,
            result=Result.CORRECT,
        )

    # endregion

    # region stimulus and reward
    def _prepare_stimulus(self) -> "NDArray[int16]":
        unit_duration = (
            self._trial_config.stimulus_freq_duration
            + self._trial_config.stimulus_freq_interval
        )
        times = ceil(self._trial_config.stimulus_duration / unit_duration)
        times = max(times, 1)

        freq_1 = ToneConfig(
            frequency=self._trial_config.stimulus_freq,
            duration=self._trial_config.stimulus_freq_duration,
        )
        freq_2 = ToneConfig(
            frequency=0,
            duration=self._trial_config.stimulus_freq_interval,
        )

        return self._theater.aplayer.generate_stimulus([freq_1, freq_2], times)

    def _set_stimulus_intensity(self) -> None:
        self._theater.acontroller.set_master_volume(self._trial_config.stimulus_density)

    def _give_stimulus(self) -> None:
        self._play_future = self._theater.aplayer.play_stimulus(self._tone)
        self._play_future.add_done_callback(self._on_stimulus_complete)

    def _direct_stimulus(self) -> None:
        self._play_future = self._theater.aplayer.play_stimulus(self._tone)
        self._play_future.add_done_callback(self._on_direct_stimulus_complete)

    def _on_direct_stimulus_complete(self, future: "Future[bool]") -> None:
        if future.result():
            self._trigger.after(0, self._give_reward)

            self._trigger.after(
                self._trial_config.reward_duration,
                lambda: (self._start_timing(), self._stimulus_loop()),
            )

    def _on_stimulus_complete(self, future: "Future[bool]") -> None:
        if future.result():
            self._trigger.after(0, self._give_reward)

            self._trigger.after(self._trial_config.reward_duration, self._stimulus_loop)

    def _on_correct(self) -> None:
        self._data.result = Result.CORRECT
        self._on_trial_end()

    def _give_reward(self) -> None:
        self._context.rewards += 1
        self._theater.reward.give_reward(self._trial_config.reward_duration)

    # endregion



from random import choice, choices, randint
from typing import TYPE_CHECKING, Final

from mxbi.data_logger import DataLogger, DataLoggerType
from mxbi.models.animal import ScheduleCondition
from mxbi.tasks.GNGSiD.models import PersistentData, Result
from mxbi.tasks.GNGSiD.stages.detect_stage.detect_stage_models import (
    DetectStageConfig,
    config,
)
from mxbi.tasks.GNGSiD.tasks.detect.models import TrialConfig
from mxbi.tasks.GNGSiD.tasks.detect.scene import GNGSiDDetectScene
from mxbi.utils.logger import logger

if TYPE_CHECKING:
    from mxbi.models.animal import AnimalState
    from mxbi.models.session import SessionState
    from mxbi.models.task import Feedback
    from mxbi.theater import Theater


_presistent_data: dict[str, PersistentData] = {}


class GNGSiDDetectStage:
    STAGE_NAME: Final[str] = "GNGSiD_DETECT_STAGE"

    def __init__(
        self,
        theater: "Theater",
        session_state: "SessionState",
        animal_state: "AnimalState",
    ) -> None:
        self._theater = theater
        self._session_state = session_state
        self._animal_state = animal_state

        self._stage_config = self._load_stage_config(animal_state.name)

        _fixed_config = self._stage_config.params
        _levels_config = self._stage_config.levels_table[animal_state.level]

        _is_go = choices(
            [True, False],
            weights=[_levels_config.go_task_prob, _levels_config.nogo_task_prob],
        )[0]

        _stimulus_duration = randint(
            _levels_config.min_stimulus_duration,
            _levels_config.max_stimulus_duration,
        )

        _master_amp, _digital_amp = self._prepare_stimulus_intensity(
            self._animal_state.name, _fixed_config.stimulus_freq
        )

        _config = TrialConfig(
            level=_levels_config.level,
            stimulation_size=_fixed_config.stimulation_size,
            stimulus_duration=_stimulus_duration,
            time_out=_fixed_config.time_out,
            inter_trial_interval=_fixed_config.inter_trial_interval,
            reward_duration=_fixed_config.reward_duration,
            reward_delay=_fixed_config.reward_delay,
            go=_is_go,
            visual_stimulus_delay=_fixed_config.visual_stimulus_delay,
            stimulus_freq=_fixed_config.stimulus_freq,
            stimulus_freq_duration=_fixed_config.stimulus_freq_duration,
            stimulus_freq_master_amp=_master_amp,
            stimulus_freq_digital_amp=_digital_amp,
            stimulus_interval=_fixed_config.stimulus_interval,
        )

        self._presistent_data = _presistent_data.get(self._animal_state.name)

        if self._presistent_data is None:
            self._presistent_data = PersistentData(
                rewards=0,
                correct=0,
                incorrect=0,
                timeout=0,
            )
            _presistent_data[self._animal_state.name] = self._presistent_data

        self._task = GNGSiDDetectScene(
            theater,
            session_state.session_config,
            animal_state,
            session_state.session_config.screen_type,
            _config,
            self._presistent_data,
        )

        self._data_logger = DataLogger(
            self._session_state,
            self._animal_state.name,
            self.STAGE_NAME,
            DataLoggerType.JSONL,
        )

    def start(self) -> "Feedback":
        trial_data = self._task.start()
        self._data_logger.save(trial_data.model_dump())

        feedback = self._handle_result(trial_data.result)
        logger.debug(
            f"{self.STAGE_NAME}: "
            f"session_id={self._session_state.session_id}, "
            f"animal_name={self._animal_state.name}, "
            f"animal_level={self._animal_state.level}, "
            f"state_name={self.STAGE_NAME}, "
            f"result={trial_data}, "
            f"feedback={feedback}"
        )

        return feedback

    def _load_stage_config(self, monkey: str) -> DetectStageConfig:
        stage_config = config.root.get(monkey) or config.root.get("default")
        if stage_config is None:
            raise ValueError("No default stage config found")
        return stage_config

    def _handle_result(self, result: "Result") -> "Feedback":
        feedback = False
        match result:
            case Result.CORRECT:
                _presistent_data[self._animal_state.name].correct += 1
                feedback = True
            case Result.INCORRECT:
                _presistent_data[self._animal_state.name].incorrect += 1
                feedback = False
            case Result.TIMEOUT:
                _presistent_data[self._animal_state.name].timeout += 1
                feedback = False
            case Result.CANCEL:
                feedback = False

        return feedback

    def quit(self) -> None:
        self._task.cancle()

    def on_idle(self) -> None:
        self._task.cancle()

    def on_return(self) -> None:
        self._task.cancle()

    @property
    def condition(self) -> "ScheduleCondition | None":
        return self._stage_config.condition

    def _prepare_stimulus_intensity(self, monkey: str, frequency: int):
        bt = choice(([[10, 30], [50, 70]])) if monkey == "wolfgang" else []
        at = [80, 80, 80] if monkey == "wolfgang" else [80, 80, 80]
        intensity_options = at * 10 + bt

        stimulus_intensity = choice(intensity_options)

        return self._theater.acontroller.get_amp_value(frequency, stimulus_intensity)



from PySide6.QtWidgets import QFrame


class CardFrame(QFrame):
    def __init__(self, parent=None, object_name: str = "card"):
        super().__init__(parent=parent)

        self.setObjectName(object_name)
        self.setFrameShape(QFrame.Shape.Box)
        self.setLineWidth(1)
        self.setStyleSheet(
            f"""
            QFrame#{object_name} {{
                border: 1px solid #C8C8C8;
                border-radius: 6px;
            }}
            """
        )



from ..scene import SceneManager
from .experiment_panel import ExperimentPanel
from .mxbi_panel import MXBIPanel


def config_wizard(scene_manager: SceneManager):
    """
    Launch the configuration wizard dialog for MXBI setup.

    Parameters
    ----------
    scene_manager : SceneManager
        The scene manager instance providing stage and level information.

    Notes
    -----
    This function creates a Qt application with a two-panel wizard:
    MXBIPanel for MXBI configuration followed by ExperimentPanel for
    experiment settings. The application exits when the wizard completes.
    """
    import sys

    from PySide6.QtWidgets import QApplication

    app = QApplication(sys.argv)

    mxbi_panel = MXBIPanel()
    experiment_panel = ExperimentPanel(scene_manager)
    mxbi_panel.accepted.connect(experiment_panel.show)
    mxbi_panel.rejected.connect(lambda: sys.exit(0))
    experiment_panel.accepted.connect(app.quit)
    experiment_panel.rejected.connect(lambda: sys.exit(0))

    mxbi_panel.show()

    app.exec()



from pathlib import Path
from typing import Any, Literal

from pydantic import BaseModel, field_validator, model_validator


class Trial(BaseModel):
    trial_id: str
    trial_number: int
    subject_id: str
    partner_id: str

    call_identity_id: str
    call_category: str
    is_partner_call: bool

    other_identity_id: str
    other_category: str

    partner_side: Literal["left", "right"]
    correct_side: Literal["left", "right"]

    audio_identity_id: str
    audio_index: int
    audio_path: str

    left_image_identity_id: str
    left_image_index: int
    left_image_path: str

    right_image_identity_id: str
    right_image_index: int
    right_image_path: str

    seed: str

    @model_validator(mode="before")
    @classmethod
    def _from_bundle_trial(cls, values: Any) -> Any:
        if not isinstance(values, dict):
            return values

        if "trialId" not in values:
            return values

        audio = values.get("audio") if isinstance(values.get("audio"), dict) else {}
        left_image = (
            values.get("leftImage") if isinstance(values.get("leftImage"), dict) else {}
        )
        right_image = (
            values.get("rightImage")
            if isinstance(values.get("rightImage"), dict)
            else {}
        )

        return {
            "trial_id": values.get("trialId"),
            "trial_number": values.get("trialNumber"),
            "subject_id": values.get("subjectId"),
            "partner_id": values.get("partnerId"),
            "call_identity_id": values.get("callIdentityId"),
            "call_category": values.get("callCategory"),
            "is_partner_call": values.get("isPartnerCall"),
            "other_identity_id": values.get("otherIdentityId"),
            "other_category": values.get("otherCategory"),
            "partner_side": values.get("partnerSide"),
            "correct_side": values.get("correctSide"),
            "audio_identity_id": audio.get("identityId"),
            "audio_index": audio.get("exemplarIndex"),
            "audio_path": audio.get("path"),
            "left_image_identity_id": left_image.get("identityId"),
            "left_image_index": left_image.get("exemplarIndex"),
            "left_image_path": left_image.get("path"),
            "right_image_identity_id": right_image.get("identityId"),
            "right_image_index": right_image.get("exemplarIndex"),
            "right_image_path": right_image.get("path"),
            "seed": values.get("seed"),
        }

    @field_validator("partner_side", "correct_side")
    @classmethod
    def _normalize_sides(cls, v: str) -> str:
        lv = v.strip().lower()
        if lv not in {"left", "right"}:
            raise ValueError("side must be 'left' or 'right'")
        return lv

    @field_validator("is_partner_call", mode="before")
    @classmethod
    def _boolify(cls, v):
        if isinstance(v, bool):
            return v
        if isinstance(v, int):
            return bool(v)
        if isinstance(v, str):
            lv = v.strip().lower()
            if lv in {"true", "t", "yes", "y", "1"}:
                return True
            if lv in {"false", "f", "no", "n", "0"}:
                return False
        raise ValueError("is_partner_call must be boolean-like")

    def audio_path_obj(self, base: Path | None = None) -> Path:
        p = Path(self.audio_path)
        return (base / p) if base and not p.is_absolute() else p

    def left_image_path_obj(self, base: Path | None = None) -> Path:
        p = Path(self.left_image_path)
        return (base / p) if base and not p.is_absolute() else p

    def right_image_path_obj(self, base: Path | None = None) -> Path:
        p = Path(self.right_image_path)
        return (base / p) if base and not p.is_absolute() else p



from typing import NamedTuple

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QFormLayout,
    QGridLayout,
    QGroupBox,
    QLabel,
    QLineEdit,
    QMenu,
)

from ...models.animal import AnimalConfig
from ...models.session import RewardEnum, SessionConfig
from .animal import AnimalCard


class ExperimentConfigGroup(QGroupBox):
    def __init__(self, parent, experimenters: list[str]):
        super().__init__("Config", parent)

        layout = QGridLayout(self)
        self.setLayout(layout)
        layout.setColumnStretch(1, 1)
        layout.setColumnStretch(3, 1)

        # Row 0: experimenter | hide cursor
        layout.addWidget(QLabel("experimenter", self), 0, 0)
        self.combo_experimenter = QComboBox(self)
        self.combo_experimenter.addItems(experimenters)
        layout.addWidget(self.combo_experimenter, 0, 1)

        layout.addWidget(QLabel("hide cursor", self), 0, 2)
        self.check_hide_cursor = QCheckBox(self)
        layout.addWidget(self.check_hide_cursor, 0, 3)

        # Row 1: reward type | fullscreen
        layout.addWidget(QLabel("reward type", self), 1, 0)
        self.combo_reward_type = QComboBox(self)
        self.combo_reward_type.addItems(list(RewardEnum))
        layout.addWidget(self.combo_reward_type, 1, 1)

        layout.addWidget(QLabel("fullscreen", self), 1, 2)
        self.check_fullscreen = QCheckBox(self)
        layout.addWidget(self.check_fullscreen, 1, 3)

        # Row 2: notes (span all columns)
        layout.addWidget(QLabel("notes", self), 2, 0)
        self.line_notes = QLineEdit(self)
        self.line_notes.setPlaceholderText("Notes")
        layout.addWidget(self.line_notes, 2, 1, 1, 3)

    def load_config(self, config: SessionConfig):
        self.combo_experimenter.setEditText(config.experimenter)
        self.combo_reward_type.setEditText(config.reward_type)
        self.check_hide_cursor.setChecked(config.hide_cursor)
        self.check_fullscreen.setChecked(config.fullscreen)

    def result(self) -> SessionConfig:
        return SessionConfig(
            experimenter=self.combo_experimenter.currentText(),
            reward_type=RewardEnum(self.combo_reward_type.currentText()),
            note=self.line_notes.text(),
            hide_cursor=self.check_hide_cursor.isChecked(),
            fullscreen=self.check_fullscreen.isChecked(),
        )


class SceneSelection(NamedTuple):
    default_scene: str
    unknown_animal_fallback: str
    fault_fallback: str


class ExperimentSceneGroup(QGroupBox):
    def __init__(self, parent=None, *, stages: list[str]):
        super().__init__("Scene", parent)

        layout = QFormLayout(self)
        self.setLayout(layout)

        label_default = QLabel("default scene", self)
        self.combo_default_scene = QComboBox(self)
        self.combo_default_scene.addItems(stages)
        layout.addRow(label_default, self.combo_default_scene)

        label_unknown = QLabel("unknown animal fallback", self)
        self.combo_unknown_animal_fallback = QComboBox(self)
        self.combo_unknown_animal_fallback.addItems(stages)
        layout.addRow(label_unknown, self.combo_unknown_animal_fallback)

        label_fault = QLabel("fault fallback", self)
        self.combo_fault_fallback = QComboBox(self)
        self.combo_fault_fallback.addItems(stages)
        layout.addRow(label_fault, self.combo_fault_fallback)

    def load_config(self, config: SessionConfig):
        if config.default_scene:
            self.combo_default_scene.setCurrentText(config.default_scene)
        if config.unknown_animal_fallback:
            self.combo_unknown_animal_fallback.setCurrentText(
                config.unknown_animal_fallback
            )
        if config.fault_fallback:
            self.combo_fault_fallback.setCurrentText(config.fault_fallback)

    def result(self) -> SceneSelection:
        return SceneSelection(
            default_scene=self.combo_default_scene.currentText(),
            unknown_animal_fallback=self.combo_unknown_animal_fallback.currentText(),
            fault_fallback=self.combo_fault_fallback.currentText(),
        )


class ExperimentAnimalsGroup(QGroupBox):
    def __init__(
        self,
        parent=None,
        *,
        animals: dict[str, str],
        stage_level_tables: dict[str, dict[str, list[int]]],
    ):
        super().__init__("Animals", parent)

        self._animals = animals
        self._stage_level_tables = stage_level_tables
        self.layout_animals = QGridLayout(self)
        self.setLayout(self.layout_animals)

        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.customContextMenuRequested.connect(self._on_context_menu)

    def _on_context_menu(self, pos):
        menu = QMenu(self)
        action = menu.addAction("Add animal")
        action.triggered.connect(lambda _checked=False: self._on_add_animal())
        menu.exec(self.mapToGlobal(pos))

    def _on_add_animal(self):
        animal_card = AnimalCard(self, self._animals, self._stage_level_tables)
        animal_card.remove_requested.connect(
            lambda _card=animal_card: self._on_remove_animal(_card)
        )
        self._add_animal_card(animal_card)

    def _add_animal_card(self, animal_card: AnimalCard) -> None:
        index = self.layout_animals.count()
        row = index // 4
        col = index % 4
        self.layout_animals.addWidget(animal_card, row, col)

    def _on_remove_animal(self, animal_card: AnimalCard) -> None:
        self.layout_animals.removeWidget(animal_card)
        animal_card.setParent(None)
        animal_card.deleteLater()
        self._reflow_animal_cards()

    def _reflow_animal_cards(self) -> None:
        cards: list[AnimalCard] = []
        for i in range(self.layout_animals.count()):
            item = self.layout_animals.itemAt(i)
            if item:
                widget = item.widget()
                if isinstance(widget, AnimalCard):
                    cards.append(widget)

        while self.layout_animals.count() > 0:
            self.layout_animals.takeAt(0)

        for card in cards:
            self._add_animal_card(card)

    def load_config(self, config: SessionConfig):
        for animal_configs in config.animals:
            animal_card = AnimalCard(self, self._animals, self._stage_level_tables)
            animal_card.load_config(animal_configs)
            animal_card.remove_requested.connect(
                lambda _card=animal_card: self._on_remove_animal(_card)
            )
            self._add_animal_card(animal_card)

    def result(self) -> list[AnimalConfig]:
        results: list[AnimalConfig] = []
        for i in range(self.layout_animals.count()):
            item = self.layout_animals.itemAt(i)
            if not item:
                continue
            widget = item.widget()
            if isinstance(widget, AnimalCard):
                results.append(widget.result)
        return results



from mxbi.models.task import TaskEnum
from mxbi.tasks.cross_modal.stage import CrossModalTask
from mxbi.tasks.default.error_task.error_scene import ErrorScene
from mxbi.tasks.default.idle_task.idle_scene import IDLEScene
from mxbi.tasks.default.initial_habituation_training.stages.initial_habituation_training_stage import (
    InitialHabituationTrainingStage,
)
from mxbi.tasks.GNGSiD.stages.detect_stage.detect_stage import GNGSiDDetectStage
from mxbi.tasks.GNGSiD.stages.discriminate_stage.discriminate_stage import (
    GNGSiDDiscriminateStage,
)
from mxbi.tasks.GNGSiD.stages.size_reduction_stage.size_reduction_stage import (
    SizeReductionStage,
)
from mxbi.tasks.task_protocol import Task
from mxbi.tasks.two_alternative_choice.stages.size_reduction_stage.size_reduction_stage import (
    TWOACSizeReductionStage,
)

task_table: dict[TaskEnum, type[Task]] = {
    TaskEnum.IDEL: IDLEScene,
    TaskEnum.ERROR: ErrorScene,
    TaskEnum.HABITUATION: InitialHabituationTrainingStage,
    TaskEnum.GNGSiD_SIZE_REDUCTION_STAGE: SizeReductionStage,
    TaskEnum.GNGSiD_DETECT_STAGE: GNGSiDDetectStage,
    TaskEnum.GNGSiD_DISCRIMINATE_STAGE: GNGSiDDiscriminateStage,
    TaskEnum.TWOAC_SIZE_REDUCTION_STAGE: TWOACSizeReductionStage,
    TaskEnum.CROSS_MODAL: CrossModalTask,
}



from datetime import datetime
from math import ceil
from tkinter import CENTER, Canvas, Event
from typing import TYPE_CHECKING, Final

from mxbi.tasks.GNGSiD.models import Result, TouchEvent
from mxbi.tasks.GNGSiD.tasks.detect.models import DataToShow, TrialConfig, TrialData
from mxbi.tasks.GNGSiD.tasks.utils.targets import DetectTarget
from mxbi.utils.aplayer import ToneConfig
from mxbi.utils.tkinter.components.canvas_with_border import CanvasWithInnerBorder
from mxbi.utils.tkinter.components.showdata_widget import ShowDataWidget

if TYPE_CHECKING:
    from concurrent.futures import Future

    from mxbi.models.animal import AnimalState
    from mxbi.models.session import ScreenConfig, SessionConfig
    from mxbi.tasks.GNGSiD.models import PersistentData
    from mxbi.theater import Theater
    from numpy import int16
    from numpy.typing import NDArray


class GNGSiDDetectScene:
    def __init__(
        self,
        theater: "Theater",
        session_config: "SessionConfig",
        animal_state: "AnimalState",
        screen_type: "ScreenConfig",
        trial_config: "TrialConfig",
        persistent_data: "PersistentData",
    ) -> None:
        self._theater: "Final[Theater]" = theater
        self._animal_state: "Final[AnimalState]" = animal_state
        self._screen_type: "Final[ScreenConfig]" = screen_type
        self._trial_config: "Final[TrialConfig]" = trial_config
        self._persistent_data: Final["PersistentData"] = persistent_data

        self._tone: Final[NDArray[int16]] = self._prepare_stimulus()
        self._standard_reward_stimulus = self._theater.new_standard_reward_stimulus(
            self._trial_config.stimulus_duration
        )

        self._set_stimulus_intensity()

        self._on_trial_start()

    # region public api
    def start(self) -> TrialData:
        self._theater.mainloop()
        return self._data

    def cancle(self) -> None:
        self._data.result = Result.CANCEL
        self._theater.aplayer.stop()
        self._on_trial_end()

    # endregion

    # region Lifecycle
    def _on_trial_start(self) -> None:
        self._create_view()
        self._init_data()
        self._bind_first_stage()

    def _on_inter_trial(self) -> None:
        self._background.after(
            self._trial_config.inter_trial_interval, self._on_trial_end
        )

    def _on_trial_end(self) -> None:
        self._background.destroy()
        self._theater.root.quit()

    # endregion

    # region Views
    def _create_view(self) -> None:
        self._create_background()
        self._create_show_data_view()
        self._create_target()

    def _create_background(self) -> None:
        self._background = CanvasWithInnerBorder(
            master=self._theater.root,
            bg="black",
            width=self._screen_type.width,
            height=self._screen_type.height,
            border_width=40,
        )
        self._background.place(relx=0.5, rely=0.5, anchor="center")

    def _create_show_data_view(self) -> None:
        self._show_data_widget = ShowDataWidget(self._background)
        self._show_data_widget.place(relx=0, rely=1, anchor="sw")
        data = DataToShow(
            name=self._animal_state.name,
            id=self._animal_state.trial_id,
            level_id=self._animal_state.current_level_trial_id,
            level=self._trial_config.level,
            rewards=self._persistent_data.rewards,
            correct=self._animal_state.correct_trial,
            incorrect=self._persistent_data.incorrect,
            timeout=self._persistent_data.timeout,
            stimulus=self._trial_config.go,
        )
        self._show_data_widget.show_data(data.model_dump())

    def _create_target(self):
        xshift = 240
        xcenter = self._screen_type.width * 0.5 + xshift
        ycenter = self._screen_type.height * 0.5

        self._trigger_canvas = DetectTarget(
            self._background, self._trial_config.stimulation_size
        )
        self._trigger_canvas.place(x=xcenter, y=ycenter, anchor="center")

    def _create_wrong_view(self) -> None:
        self._trigger_canvas = Canvas(
            self._background,
            bg="grey",
            width=self._screen_type.width,
            height=self._screen_type.height,
        )
        self._trigger_canvas.place(relx=0.5, rely=0.5, anchor=CENTER)

    # endregion

    # region event binding
    def _bind_first_stage(self) -> None:
        self._background.focus_set()
        self._background.bind("<r>", lambda e: self._give_standard_stimulus())
        self._trigger_canvas.bind("<ButtonPress>", self._on_first_touched)
        self._trigger_canvas.after(self._trial_config.time_out, self._on_timeout)

    def _bind_second_stage(self) -> None:
        self._trigger_canvas.bind("<ButtonPress>", self._on_second_touched)

        # TODO: Confirm the waiting time
        if not self._trial_config.go:
            self._trigger_canvas.after(
                self._trial_config.stimulus_duration, self._on_incorrect
            )

    # endregion

    # region event handlers
    def _on_first_touched(self, event: Event) -> None:
        self._trigger_canvas.destroy()
        self._record_touch(event)

        if self._trial_config.go:
            future = self._give_stimulus(self._tone)
            future.add_done_callback(self._on_stimulus_complete)

        self._background.after(
            self._trial_config.visual_stimulus_delay,
            lambda: (self._create_target(), self._bind_second_stage()),
        )

    def _on_second_touched(self, event: Event) -> None:
        self._trigger_canvas.destroy()
        self._record_touch(event)

        if self._trial_config.go:
            self._on_incorrect()
        else:
            future = self._give_stimulus(self._tone)
            future.add_done_callback(self._on_stimulus_complete)

        self._background.after(2000, self._create_target)

    def _record_touch(self, event: Event) -> None:
        self._data.touch_events.append(
            TouchEvent(time=datetime.now().timestamp(), x=event.x, y=event.y)
        )

    # endregion

    # region result handlers
    def _on_correct(self) -> None:
        self._trigger_canvas.destroy()

        self._give_reward()
        self._data.result = Result.CORRECT
        self._data.correct_rate = (self._animal_state.correct_trial + 1) / (
            self._animal_state.current_level_trial_id + 1
        )
        self._on_inter_trial()

    def _on_incorrect(self) -> None:
        self._theater._aplayer.stop()
        self._trigger_canvas.destroy()

        self._data.result = Result.INCORRECT
        self._data.correct_rate = self._animal_state.correct_trial / (
            self._animal_state.current_level_trial_id + 1
        )

        self._create_wrong_view()
        self._on_inter_trial()

    def _on_timeout(self) -> None:
        self._data.result = Result.TIMEOUT
        self._data.correct_rate = self._animal_state.correct_trial / (
            self._animal_state.current_level_trial_id + 1
        )

        self._create_wrong_view()
        self._on_inter_trial()

    # endregion

    # region stimulus and reward
    def _prepare_stimulus(self) -> "NDArray[int16]":
        cycle = (
            self._trial_config.stimulus_freq_duration
            + self._trial_config.stimulus_interval
        )
        repeat = ceil(self._trial_config.stimulus_duration / cycle)
        repeat = max(repeat, 1)

        freq_1 = ToneConfig(
            frequency=self._trial_config.stimulus_freq,
            duration=self._trial_config.stimulus_freq_duration,
        )
        freq_2 = ToneConfig(frequency=0, duration=self._trial_config.stimulus_interval)

        return self._theater.aplayer.generate_stimulus([freq_1, freq_2], repeat)

    def _give_stimulus(self, tone: "NDArray[int16]") -> "Future[bool]":
        return self._theater.aplayer.play_stimulus(tone)

    def _on_stimulus_complete(self, future: "Future[bool]") -> None:
        if future.result():
            self._background.after(self._trial_config.reward_delay, self._on_correct)

    def _give_reward(self, _=None) -> None:
        self._persistent_data.rewards += 1
        self._theater.reward.give_reward(self._trial_config.reward_duration)

    def _set_stimulus_intensity(self) -> None:
        self._theater.acontroller.set_master_volume(
            self._trial_config.stimulus_freq_master_amp
        )
        self._theater.acontroller.set_digital_volume(
            self._trial_config.stimulus_freq_digital_amp
        )

    def _give_standard_stimulus(self) -> None:
        self._standard_reward_stimulus.play(self._trial_config.reward_duration)

    # endregion

    # region data
    def _init_data(self):
        self._data = TrialData(
            animal=self._animal_state.name,
            trial_id=self._animal_state.trial_id,
            current_level_trial_id=self._animal_state.current_level_trial_id,
            trial_config=self._trial_config,
            trial_start_time=datetime.now().timestamp(),
            trial_end_time=0,
            result=Result.TIMEOUT,
            correct_rate=0,
            touch_events=[],
        )

    # endregion
